"""WS API for common platform entity functionality."""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Optional

import voluptuous as vol

from zhaws.backports.enum import StrEnum
from zhaws.server.const import ATTR_UNIQUE_ID, COMMAND, IEEE
from zhaws.server.websocket.api import decorators, register_api_command

if TYPE_CHECKING:
    from zhaws.server.websocket.client import Client
    from zhaws.server.websocket.server import Server

_LOGGER = logging.getLogger(__name__)


class PlatformEntityCommands(StrEnum):
    """Enumeration of platform entity commands."""

    REFRESH_STATE = "platform_entity_refresh_state"


def platform_entity_command_schema(
    command: str, schema: Optional[dict[str, Any]] = None
) -> dict[str, Any]:
    """Return the schema for a platform entity command."""
    full_schema = {
        vol.Required(COMMAND): str(command),
        vol.Optional(IEEE): str,
        vol.Optional("group_id"): int,
        vol.Required(ATTR_UNIQUE_ID): str,
    }
    if schema:
        full_schema.update(schema)
    return full_schema


async def execute_platform_entity_command(
    server: Server,
    client: Client,
    request_message: dict[str, Any],
    command: str,
) -> None:
    """Get the platform entity and execute a command."""
    try:
        if IEEE in request_message:
            device = server.controller.get_device(request_message[IEEE])
            platform_entity: Any = device.get_platform_entity(
                request_message[ATTR_UNIQUE_ID]
            )
        else:
            group = server.controller.get_group(request_message["group_id"])
            platform_entity = group.group_entities[request_message[ATTR_UNIQUE_ID]]
    except ValueError as err:
        _LOGGER.exception("Error executing command: %s", command, exc_info=err)
        client.send_result_error(
            request_message, "PLATFORM_ENTITY_COMMAND_ERROR", str(err)
        )
        return None

    try:
        action = getattr(platform_entity, command)
        if action.__code__.co_argcount == 1:  # the only argument is self
            await action()
        else:
            await action(**request_message)
    except Exception as err:
        _LOGGER.exception("Error executing command: %s", command, exc_info=err)
        client.send_result_error(
            request_message, "PLATFORM_ENTITY_ACTION_ERROR", str(err)
        )

    result = {}
    if IEEE in request_message:
        result[IEEE] = request_message[IEEE]
    else:
        result["group_id"] = request_message["group_id"]
    result[ATTR_UNIQUE_ID] = request_message[ATTR_UNIQUE_ID]
    client.send_result_success(request_message, result)


@decorators.websocket_command(
    platform_entity_command_schema(PlatformEntityCommands.REFRESH_STATE)
)
@decorators.async_response
async def refresh_state(
    server: Server, client: Client, message: dict[str, Any]
) -> None:
    """Refresh the state of the platform entity."""
    await execute_platform_entity_command(server, client, message, "async_update")


def load_platform_entity_apis(server: Server) -> None:
    """Load the ws apis for all platform entities types."""
    from zhaws.server.platforms.alarm_control_panel.api import (
        load_api as load_alarm_control_panel_api,
    )
    from zhaws.server.platforms.button.api import load_api as load_button_api
    from zhaws.server.platforms.climate.api import load_api as load_climate_api
    from zhaws.server.platforms.cover.api import load_api as load_cover_api
    from zhaws.server.platforms.fan.api import load_api as load_fan_api
    from zhaws.server.platforms.light.api import load_api as load_light_api
    from zhaws.server.platforms.lock.api import load_api as load_lock_api
    from zhaws.server.platforms.number.api import load_api as load_number_api
    from zhaws.server.platforms.select.api import load_api as load_select_api
    from zhaws.server.platforms.siren.api import load_api as load_siren_api
    from zhaws.server.platforms.switch.api import load_api as load_switch_api

    register_api_command(server, refresh_state)
    load_alarm_control_panel_api(server)
    load_button_api(server)
    load_climate_api(server)
    load_cover_api(server)
    load_fan_api(server)
    load_light_api(server)
    load_lock_api(server)
    load_number_api(server)
    load_select_api(server)
    load_siren_api(server)
    load_switch_api(server)



"""Websocket application to run a zigpy Zigbee network."""

import argparse
import asyncio
import logging

from zhaws.server.websocket.server import Server

_LOGGER = logging.getLogger(__name__)


async def main(host: str, port: int) -> None:
    async with Server(host=host, port=port) as server:
        await server.wait_closed()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Start the ZHAWS server")
    parser.add_argument("--host", type=str, default="0.0.0.0", help="Bind host")
    parser.add_argument("--port", type=int, default=8001, help="Bind port")

    args = parser.parse_args()

    import uvloop

    uvloop.install()

    from colorlog import ColoredFormatter

    fmt = "%(asctime)s %(levelname)s (%(threadName)s) [%(name)s] %(message)s"
    colorfmt = f"%(log_color)s{fmt}%(reset)s"
    logging.basicConfig(level=logging.DEBUG)
    logging.getLogger().handlers[0].setFormatter(
        ColoredFormatter(
            colorfmt,
            reset=True,
            log_colors={
                "DEBUG": "cyan",
                "INFO": "green",
                "WARNING": "yellow",
                "ERROR": "red",
                "CRITICAL": "red",
            },
        )
    )

    asyncio.run(main(args.host, args.port))



"""
Security cluster handlers module for zhawss.

For more details about this component, please refer to the documentation at
https://home-assistant.io/integrations/zha/
"""
from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING, Any, Callable, Literal

from zigpy.exceptions import ZigbeeException
from zigpy.zcl import Cluster as ZigpyClusterType
from zigpy.zcl.clusters import security
from zigpy.zcl.clusters.security import IasAce as AceCluster

from zhaws.model import BaseEvent
from zhaws.server.zigbee import registries
from zhaws.server.zigbee.cluster import (
    CLUSTER_HANDLER_EVENT,
    ClusterHandler,
    ClusterHandlerStatus,
)

if TYPE_CHECKING:
    from zhaws.server.zigbee.endpoint import Endpoint

IAS_ACE_ARM = 0x0000  # ("arm", (t.enum8, t.CharacterString, t.uint8_t), False),
IAS_ACE_BYPASS = 0x0001  # ("bypass", (t.LVList(t.uint8_t), t.CharacterString), False),
IAS_ACE_EMERGENCY = 0x0002  # ("emergency", (), False),
IAS_ACE_FIRE = 0x0003  # ("fire", (), False),
IAS_ACE_PANIC = 0x0004  # ("panic", (), False),
IAS_ACE_GET_ZONE_ID_MAP = 0x0005  # ("get_zone_id_map", (), False),
IAS_ACE_GET_ZONE_INFO = 0x0006  # ("get_zone_info", (t.uint8_t,), False),
IAS_ACE_GET_PANEL_STATUS = 0x0007  # ("get_panel_status", (), False),
IAS_ACE_GET_BYPASSED_ZONE_LIST = 0x0008  # ("get_bypassed_zone_list", (), False),
IAS_ACE_GET_ZONE_STATUS = (
    0x0009  # ("get_zone_status", (t.uint8_t, t.uint8_t, t.Bool, t.bitmap16), False)
)
NAME = 0
SIGNAL_ARMED_STATE_CHANGED = "zha_armed_state_changed"
SIGNAL_ALARM_TRIGGERED = "zha_armed_triggered"

WARNING_DEVICE_MODE_STOP = 0
WARNING_DEVICE_MODE_BURGLAR = 1
WARNING_DEVICE_MODE_FIRE = 2
WARNING_DEVICE_MODE_EMERGENCY = 3
WARNING_DEVICE_MODE_POLICE_PANIC = 4
WARNING_DEVICE_MODE_FIRE_PANIC = 5
WARNING_DEVICE_MODE_EMERGENCY_PANIC = 6

WARNING_DEVICE_STROBE_NO = 0
WARNING_DEVICE_STROBE_YES = 1

WARNING_DEVICE_SOUND_LOW = 0
WARNING_DEVICE_SOUND_MEDIUM = 1
WARNING_DEVICE_SOUND_HIGH = 2
WARNING_DEVICE_SOUND_VERY_HIGH = 3

WARNING_DEVICE_STROBE_LOW = 0x00
WARNING_DEVICE_STROBE_MEDIUM = 0x01
WARNING_DEVICE_STROBE_HIGH = 0x02
WARNING_DEVICE_STROBE_VERY_HIGH = 0x03

WARNING_DEVICE_SQUAWK_MODE_ARMED = 0
WARNING_DEVICE_SQUAWK_MODE_DISARMED = 1

_LOGGER = logging.getLogger(__name__)


class ClusterHandlerStateChangedEvent(BaseEvent):
    """Event to signal that a cluster attribute has been updated."""

    event_type: Literal["cluster_handler_event"] = "cluster_handler_event"
    event: Literal["cluster_handler_state_changed"] = "cluster_handler_state_changed"


@registries.CLUSTER_HANDLER_REGISTRY.register(AceCluster.cluster_id)
class IasAce(ClusterHandler):
    """IAS Ancillary Control Equipment cluster handler."""

    def __init__(self, cluster: ZigpyClusterType, endpoint: Endpoint) -> None:
        """Initialize IAS Ancillary Control Equipment cluster handler."""
        super().__init__(cluster, endpoint)
        self.command_map: dict[int, Callable] = {
            IAS_ACE_ARM: self.arm,
            IAS_ACE_BYPASS: self._bypass,
            IAS_ACE_EMERGENCY: self._emergency,
            IAS_ACE_FIRE: self._fire,
            IAS_ACE_PANIC: self._panic,
            IAS_ACE_GET_ZONE_ID_MAP: self._get_zone_id_map,
            IAS_ACE_GET_ZONE_INFO: self._get_zone_info,
            IAS_ACE_GET_PANEL_STATUS: self._send_panel_status_response,
            IAS_ACE_GET_BYPASSED_ZONE_LIST: self._get_bypassed_zone_list,
            IAS_ACE_GET_ZONE_STATUS: self._get_zone_status,
        }
        self.arm_map: dict[AceCluster.ArmMode, Callable] = {
            AceCluster.ArmMode.Disarm: self._disarm,
            AceCluster.ArmMode.Arm_All_Zones: self._arm_away,
            AceCluster.ArmMode.Arm_Day_Home_Only: self._arm_day,
            AceCluster.ArmMode.Arm_Night_Sleep_Only: self._arm_night,
        }
        self.armed_state: AceCluster.PanelStatus = AceCluster.PanelStatus.Panel_Disarmed
        self.invalid_tries: int = 0

        # These will all be setup by the entity from zha configuration
        self.panel_code: str = "1234"
        self.code_required_arm_actions: bool = False
        self.max_invalid_tries: int = 3

        # where do we store this to handle restarts
        self.alarm_status: AceCluster.AlarmStatus = AceCluster.AlarmStatus.No_Alarm

    def cluster_command(self, tsn: int, command_id: int, args: Any) -> None:
        """Handle commands received to this cluster."""
        self.warning(
            "received command %s", self._cluster.server_commands.get(command_id)[NAME]
        )
        self.command_map[command_id](*args)

    def arm(self, arm_mode: int, code: str | None, zone_id: int) -> None:
        """Handle the IAS ACE arm command."""
        mode = AceCluster.ArmMode(arm_mode)

        """TODO figure out events
        self.zha_send_event(
            self._cluster.server_commands.get(IAS_ACE_ARM)[NAME],
            {
                "arm_mode": mode.value,
                "arm_mode_description": mode.name,
                "code": code,
                "zone_id": zone_id,
            },
        )
        """

        zigbee_reply = self.arm_map[mode](code)
        asyncio.create_task(zigbee_reply)

        if self.invalid_tries >= self.max_invalid_tries:
            self.alarm_status = AceCluster.AlarmStatus.Emergency
            self.armed_state = AceCluster.PanelStatus.In_Alarm
        self._send_panel_status_changed()

    def _disarm(self, code: str) -> asyncio.Future:
        """Test the code and disarm the panel if the code is correct."""
        if (
            code != self.panel_code
            and self.armed_state != AceCluster.PanelStatus.Panel_Disarmed
        ):
            self.warning("Invalid code supplied to IAS ACE")
            self.invalid_tries += 1
            zigbee_reply = self.arm_response(
                AceCluster.ArmNotification.Invalid_Arm_Disarm_Code
            )
        else:
            self.invalid_tries = 0
            if (
                self.armed_state == AceCluster.PanelStatus.Panel_Disarmed
                and self.alarm_status == AceCluster.AlarmStatus.No_Alarm
            ):
                self.warning("IAS ACE already disarmed")
                zigbee_reply = self.arm_response(
                    AceCluster.ArmNotification.Already_Disarmed
                )
            else:
                self.warning("Disarming all IAS ACE zones")
                zigbee_reply = self.arm_response(
                    AceCluster.ArmNotification.All_Zones_Disarmed
                )

            self.armed_state = AceCluster.PanelStatus.Panel_Disarmed
            self.alarm_status = AceCluster.AlarmStatus.No_Alarm
        return zigbee_reply

    def _arm_day(self, code: str) -> asyncio.Future:
        """Arm the panel for day / home zones."""
        return self._handle_arm(
            code,
            AceCluster.PanelStatus.Armed_Stay,
            AceCluster.ArmNotification.Only_Day_Home_Zones_Armed,
        )

    def _arm_night(self, code: str) -> asyncio.Future:
        """Arm the panel for night / sleep zones."""
        return self._handle_arm(
            code,
            AceCluster.PanelStatus.Armed_Night,
            AceCluster.ArmNotification.Only_Night_Sleep_Zones_Armed,
        )

    def _arm_away(self, code: str) -> asyncio.Future:
        """Arm the panel for away mode."""
        return self._handle_arm(
            code,
            AceCluster.PanelStatus.Armed_Away,
            AceCluster.ArmNotification.All_Zones_Armed,
        )

    def _handle_arm(
        self,
        code: str,
        panel_status: AceCluster.PanelStatus,
        armed_type: AceCluster.ArmNotification,
    ) -> asyncio.Future:
        """Arm the panel with the specified statuses."""
        if self.code_required_arm_actions and code != self.panel_code:
            self.warning("Invalid code supplied to IAS ACE")
            zigbee_reply = self.arm_response(
                AceCluster.ArmNotification.Invalid_Arm_Disarm_Code
            )
        else:
            self.warning("Arming all IAS ACE zones")
            self.armed_state = panel_status
            zigbee_reply = self.arm_response(armed_type)
        return zigbee_reply

    def _bypass(self, zone_list: Any, code: str) -> asyncio.Future:
        """Handle the IAS ACE bypass command."""
        """TODO figure out events
        self.zha_send_event(
            self._cluster.server_commands.get(IAS_ACE_BYPASS)[NAME],
            {"zone_list": zone_list, "code": code},
        )
        """

    def _emergency(self) -> asyncio.Future:
        """Handle the IAS ACE emergency command."""
        self._set_alarm(AceCluster.AlarmStatus.Emergency)

    def _fire(self) -> asyncio.Future:
        """Handle the IAS ACE fire command."""
        self._set_alarm(AceCluster.AlarmStatus.Fire)

    def _panic(self) -> asyncio.Future:
        """Handle the IAS ACE panic command."""
        self._set_alarm(AceCluster.AlarmStatus.Emergency_Panic)

    def _set_alarm(self, status: AceCluster.AlarmStatus) -> asyncio.Future:
        """Set the specified alarm status."""
        self.alarm_status = status
        self.armed_state = AceCluster.PanelStatus.In_Alarm
        self.emit(
            CLUSTER_HANDLER_EVENT,
            ClusterHandlerStateChangedEvent(),
        )
        self._send_panel_status_changed()

    def _get_zone_id_map(self) -> asyncio.Future:
        """Handle the IAS ACE zone id map command."""

    def _get_zone_info(self, zone_id: int) -> asyncio.Future:
        """Handle the IAS ACE zone info command."""

    def _send_panel_status_response(self) -> asyncio.Future:
        """Handle the IAS ACE panel status response command."""
        response = self.panel_status_response(
            self.armed_state,
            0x00,
            AceCluster.AudibleNotification.Default_Sound,
            self.alarm_status,
        )
        asyncio.create_task(response)

    def _send_panel_status_changed(self) -> asyncio.Future:
        """Handle the IAS ACE panel status changed command."""
        response = self.panel_status_changed(
            self.armed_state,
            0x00,
            AceCluster.AudibleNotification.Default_Sound,
            self.alarm_status,
        )
        asyncio.create_task(response)
        self.emit(
            CLUSTER_HANDLER_EVENT,
            ClusterHandlerStateChangedEvent(),
        )

    def _get_bypassed_zone_list(self) -> asyncio.Future:
        """Handle the IAS ACE bypassed zone list command."""

    def _get_zone_status(
        self,
        starting_zone_id: int,
        max_zone_ids: int,
        zone_status_mask_flag: int,
        zone_status_mask: int,
    ) -> asyncio.Future:
        """Handle the IAS ACE zone status command."""


@registries.HANDLER_ONLY_CLUSTERS.register(security.IasWd.cluster_id)
@registries.CLUSTER_HANDLER_REGISTRY.register(security.IasWd.cluster_id)
class IasWd(ClusterHandler):
    """IAS Warning Device cluster handler."""

    @staticmethod
    def set_bit(
        destination_value: int, destination_bit: int, source_value: int, source_bit: int
    ) -> int:
        """Set the specified bit in the value."""

        if IasWd.get_bit(source_value, source_bit):
            return destination_value | (1 << destination_bit)
        return destination_value

    @staticmethod
    def get_bit(value: int, bit: int) -> bool:
        """Get the specified bit from the value."""
        return (value & (1 << bit)) != 0

    async def issue_squawk(
        self,
        mode: int = WARNING_DEVICE_SQUAWK_MODE_ARMED,
        strobe: int = WARNING_DEVICE_STROBE_YES,
        squawk_level: int = WARNING_DEVICE_SOUND_HIGH,
    ) -> None:
        """Issue a squawk command.

        This command uses the WD capabilities to emit a quick audible/visible pulse called a
        "squawk". The squawk command has no effect if the WD is currently active
        (warning in progress).
        """
        value = 0
        value = IasWd.set_bit(value, 0, squawk_level, 0)
        value = IasWd.set_bit(value, 1, squawk_level, 1)

        value = IasWd.set_bit(value, 3, strobe, 0)

        value = IasWd.set_bit(value, 4, mode, 0)
        value = IasWd.set_bit(value, 5, mode, 1)
        value = IasWd.set_bit(value, 6, mode, 2)
        value = IasWd.set_bit(value, 7, mode, 3)

        await self.squawk(value)

    async def issue_start_warning(
        self,
        mode: int = WARNING_DEVICE_MODE_EMERGENCY,
        strobe: int = WARNING_DEVICE_STROBE_YES,
        siren_level: int = WARNING_DEVICE_SOUND_HIGH,
        warning_duration: int = 5,  # seconds
        strobe_duty_cycle: int = 0x00,
        strobe_intensity: int = WARNING_DEVICE_STROBE_HIGH,
    ) -> None:
        """Issue a start warning command.

        This command starts the WD operation. The WD alerts the surrounding area by audible
        (siren) and visual (strobe) signals.

        strobe_duty_cycle indicates the length of the flash cycle. This provides a means
        of varying the flash duration for different alarm types (e.g., fire, police, burglar).
        Valid range is 0-100 in increments of 10. All other values SHALL be rounded to the
        nearest valid value. Strobe SHALL calculate duty cycle over a duration of one second.
        The ON state SHALL precede the OFF state. For example, if Strobe Duty Cycle Field specifies
        “40,” then the strobe SHALL flash ON for 4/10ths of a second and then turn OFF for
        6/10ths of a second.
        """
        value = 0
        value = IasWd.set_bit(value, 0, siren_level, 0)
        value = IasWd.set_bit(value, 1, siren_level, 1)

        value = IasWd.set_bit(value, 2, strobe, 0)

        value = IasWd.set_bit(value, 4, mode, 0)
        value = IasWd.set_bit(value, 5, mode, 1)
        value = IasWd.set_bit(value, 6, mode, 2)
        value = IasWd.set_bit(value, 7, mode, 3)

        await self.start_warning(
            value, warning_duration, strobe_duty_cycle, strobe_intensity
        )


@registries.CLUSTER_HANDLER_REGISTRY.register(security.IasZone.cluster_id)
class IASZoneClusterHandler(ClusterHandler):
    """Cluster handler for the IASZone Zigbee cluster."""

    ZCL_INIT_ATTRS = {"zone_status": True, "zone_state": False, "zone_type": True}

    def cluster_command(self, tsn: int, command_id: int, args: Any) -> None:
        """Handle commands received to this cluster."""
        _LOGGER.info("received cluster_command: %s args: %s", command_id, args)
        if command_id == 0:
            self.attribute_updated(2, args[0])
        elif command_id == 1:
            self.debug("Enroll requested")
            res = self._cluster.enroll_response(0, 0)
            asyncio.create_task(res)

    async def async_configure(self) -> None:
        """Configure IAS device."""
        await self.get_attribute_value("zone_type", from_cache=False)
        if self._endpoint.device.skip_configuration:
            self.debug("skipping IASZoneClusterHandler configuration")
            return

        self.debug("started IASZoneClusterHandler configuration")

        await self.bind()
        ieee = self.cluster.endpoint.device.application.ieee

        try:
            res = await self._cluster.write_attributes({"cie_addr": ieee})
            self.debug(
                "wrote cie_addr: %s to '%s' cluster: %s",
                str(ieee),
                self._cluster.ep_attribute,
                res[0],
            )
        except ZigbeeException as ex:
            self.debug(
                "Failed to write cie_addr: %s to '%s' cluster: %s",
                str(ieee),
                self._cluster.ep_attribute,
                str(ex),
            )

        self.debug("Sending pro-active IAS enroll response")
        self._cluster.create_catching_task(self._cluster.enroll_response(0, 0))

        self._status = ClusterHandlerStatus.CONFIGURED
        self.debug("finished IASZoneClusterHandler configuration")

    def attribute_updated(self, attrid: int, value: Any) -> None:
        """Handle attribute updates on this cluster."""
        if attrid == 2:
            value = value & 3
            super().attribute_updated(attrid, value)



"""Shared models for zhaws."""

from typing import Literal

from pydantic import BaseModel as PydanticBaseModel


class BaseModel(PydanticBaseModel):
    """Base model for zhawss models."""

    class Config:
        arbitrary_types_allowed = True
        extra = "allow"


class BaseEvent(BaseModel):
    """Base event model."""

    message_type: Literal["event"] = "event"
    event_type: str
    event: str



"""WS API for the button platform entity."""
from __future__ import annotations

from typing import TYPE_CHECKING, Any

from zhaws.server.platforms.api import (
    execute_platform_entity_command,
    platform_entity_command_schema,
)
from zhaws.server.websocket.api import decorators, register_api_command

if TYPE_CHECKING:
    from zhaws.server.websocket.client import Client
    from zhaws.server.websocket.server import Server

COMMAND_PRESS = "button_press"


@decorators.websocket_command(platform_entity_command_schema(COMMAND_PRESS))
@decorators.async_response
async def press(server: Server, client: Client, message: dict[str, Any]) -> None:
    """Turn on the button."""
    await execute_platform_entity_command(server, client, message, "async_press")


def load_api(server: Server) -> None:
    """Load the api command handlers."""
    register_api_command(server, press)



"""Client implementation for the zhaws.client."""
from __future__ import annotations

import asyncio
import contextlib
import logging
import pprint
from types import TracebackType
from typing import Any, Optional

from aiohttp import ClientSession, ClientWebSocketResponse, client_exceptions
from aiohttp.http_websocket import WSMsgType

from zhaws.client.model.commands import CommandResponse
from zhaws.client.model.messages import Message
from zhaws.event import EventBase

SIZE_PARSE_JSON_EXECUTOR = 8192
_LOGGER = logging.getLogger(__package__)


class Client(EventBase):
    """Class to manage the IoT connection."""

    def __init__(
        self,
        ws_server_url: str,
        aiohttp_session: ClientSession | None = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Initialize the Client class."""
        super().__init__(*args, **kwargs)
        self.ws_server_url = ws_server_url

        # Create a session if none is provided
        if aiohttp_session is None:
            self.aiohttp_session = ClientSession()
            self._close_aiohttp_session: bool = True
        else:
            self.aiohttp_session = aiohttp_session
            self._close_aiohttp_session: bool = False

        # The WebSocket client
        self._client: Optional[ClientWebSocketResponse] = None
        self._loop = asyncio.get_running_loop()
        self._result_futures: dict[int, asyncio.Future] = {}
        self._listen_task: asyncio.Task | None = None

        self._message_id = 0

    def __repr__(self) -> str:
        """Return the representation."""
        prefix = "" if self.connected else "not "
        return f"{type(self).__name__}(ws_server_url={self.ws_server_url!r}, {prefix}connected)"

    @property
    def connected(self) -> bool:
        """Return if we're currently connected."""
        return self._client is not None and not self._client.closed

    def new_message_id(self) -> int:
        """Creates a new message ID."""
        # XXX: JSON doesn't define limits for integers but JavaScript itself internally
        # uses double precision floats for numbers (including in `JSON.parse`), setting
        # a hard limit of `Number.MAX_SAFE_INTEGER == 2^53 - 1`.  We can be more
        # conservative and just restrict it to the maximum value of a 32-bit signed int.
        self._message_id = (self._message_id + 1) % 0x80000000
        return self._message_id

    async def async_send_command(
        self,
        message: dict[str, Any],
    ) -> CommandResponse:
        """Send a command and get a response."""
        future: asyncio.Future[CommandResponse] = self._loop.create_future()
        message_id = message["message_id"] = self.new_message_id()
        self._result_futures[message_id] = future

        try:
            await self._send_json_message(message)
            return await future
        finally:
            self._result_futures.pop(message_id)

    async def async_send_command_no_wait(self, message: dict[str, Any]) -> None:
        """Send a command without waiting for the response."""
        message["message_id"] = self.new_message_id()
        await self._send_json_message(message)

    async def connect(self) -> None:
        """Connect to the websocket server."""

        _LOGGER.debug("Trying to connect")
        try:
            self._client = await self.aiohttp_session.ws_connect(
                self.ws_server_url,
                heartbeat=55,
                compress=15,
                max_msg_size=0,
            )
        except client_exceptions.ClientError as err:
            _LOGGER.error("Error connecting to server: %s", err)
            raise err

    async def listen_loop(self) -> None:
        while not self._client.closed:
            data = await self._receive_json_or_raise()
            self._handle_incoming_message(data)

    async def listen(self) -> None:
        """Start listening to the websocket."""
        if not self.connected:
            raise Exception("Not connected when start listening")

        assert self._client

        assert self._listen_task is None
        self._listen_task = asyncio.create_task(self.listen_loop())

    async def disconnect(self) -> None:
        """Disconnect the client."""
        _LOGGER.debug("Closing client connection")

        if self._listen_task is not None:
            self._listen_task.cancel()

            with contextlib.suppress(asyncio.CancelledError):
                await self._listen_task

            self._listen_task = None

        await self._client.close()

        if self._close_aiohttp_session:
            await self.aiohttp_session.close()

        _LOGGER.debug("Listen completed. Cleaning up")

        for future in self._result_futures.values():
            future.cancel()

        self._result_futures.clear()

    async def _receive_json_or_raise(self) -> dict:
        """Receive json or raise."""
        assert self._client
        msg = await self._client.receive()

        if msg.type in (WSMsgType.CLOSE, WSMsgType.CLOSED, WSMsgType.CLOSING):
            raise Exception("Connection was closed.")

        if msg.type == WSMsgType.ERROR:
            raise Exception()

        if msg.type != WSMsgType.TEXT:
            raise Exception(f"Received non-Text message: {msg.type}")

        try:
            if len(msg.data) > SIZE_PARSE_JSON_EXECUTOR:
                data: dict = await self._loop.run_in_executor(None, msg.json)
            else:
                data = msg.json()
        except ValueError as err:
            raise Exception("Received invalid JSON.") from err

        if _LOGGER.isEnabledFor(logging.DEBUG):
            _LOGGER.debug("Received message:\n%s\n", pprint.pformat(msg))

        return data

    def _handle_incoming_message(self, msg: dict) -> None:
        """Handle incoming message.

        Run all async tasks in a wrapper to log appropriately.
        """

        try:
            message = Message.parse_obj(msg).__root__
        except Exception as err:
            _LOGGER.error("Error parsing message: %s", msg, exc_info=err)

        if message.message_type == "result":
            future = self._result_futures.get(message.message_id)

            if future is None:
                # no listener for this result
                return

            if message.success:
                future.set_result(message)
                return

            if msg["error_code"] != "zigbee_error":
                error = Exception(msg["message_id"], msg["error_code"])
            else:
                error = Exception(
                    msg["message_id"],
                    msg["zigbee_error_code"],
                    msg["zigbee_error_message"],
                )

            future.set_exception(error)
            return

        if message.message_type != "event":
            # Can't handle
            _LOGGER.debug(  # type: ignore #TODO why does mypy thins this is unreachable
                "Received message with unknown type '%s': %s",
                msg["message_type"],
                msg,
            )
            return

        try:
            self.emit(message.event_type, message)
        except Exception as err:
            _LOGGER.error("Error handling event: %s", err, exc_info=err)

    async def _send_json_message(self, message: dict[str, Any]) -> None:
        """Send a message.

        Raises NotConnected if client not connected.
        """
        if not self.connected:
            raise Exception()

        _LOGGER.warning("Publishing message:\n%s\n", pprint.pformat(message))

        assert self._client
        assert "message_id" in message

        await self._client.send_json(message)

    async def __aenter__(self) -> Client:
        """Connect to the websocket."""
        await self.connect()
        return self

    async def __aexit__(
        self, exc_type: Exception, exc_value: str, traceback: TracebackType
    ) -> None:
        """Disconnect from the websocket."""
        await self.disconnect()



"""Decorators for the Websocket API."""
from __future__ import annotations

import asyncio
from collections.abc import Callable
from functools import wraps
import logging
from typing import TYPE_CHECKING, Any

import voluptuous as vol

from zhaws.server.const import COMMAND, MINIMAL_MESSAGE_SCHEMA

if TYPE_CHECKING:
    from zhaws.server.websocket.api.types import (
        AsyncWebSocketCommandHandler,
        WebSocketCommandHandler,
    )
    from zhaws.server.websocket.client import Client
    from zhaws.server.websocket.server import Server

POSITIVE_INT = vol.All(vol.Coerce(int), vol.Range(min=0))
_LOGGER = logging.getLogger(__name__)


async def _handle_async_response(
    func: AsyncWebSocketCommandHandler,
    server: Server,
    client: Client,
    msg: dict[str, Any],
) -> None:
    """Create a response and handle exception."""
    try:
        await func(server, client, msg)
    except Exception as err:  # pylint: disable=broad-except
        # TODO fix this to send a real error code and message
        _LOGGER.exception("Error handling message: %s", err, exc_info=err)
        client.send_result_error(msg, "API_COMMAND_HANDLER_ERROR", str(err))


def async_response(
    func: AsyncWebSocketCommandHandler,
) -> WebSocketCommandHandler:
    """Decorate an async function to handle WebSocket API messages."""

    @wraps(func)
    def schedule_handler(server: Server, client: Client, msg: dict[str, Any]) -> None:
        """Schedule the handler."""
        # As the webserver is now started before the start
        # event we do not want to block for websocket responders
        asyncio.create_task(_handle_async_response(func, server, client, msg))

    return schedule_handler


def websocket_command(
    schema: dict[vol.Marker, Any],
) -> Callable[[WebSocketCommandHandler], WebSocketCommandHandler]:
    """Tag a function as a websocket command."""
    command = schema[COMMAND]

    def decorate(func: WebSocketCommandHandler) -> WebSocketCommandHandler:
        """Decorate ws command function."""
        # pylint: disable=protected-access
        func._ws_schema = MINIMAL_MESSAGE_SCHEMA.extend(schema)  # type: ignore[attr-defined]
        func._ws_command = command  # type: ignore[attr-defined]
        return func

    return decorate



"""Base classes for zigbee cluster handlers."""
from __future__ import annotations

import asyncio
from enum import Enum
from functools import partialmethod
import logging
from typing import TYPE_CHECKING, Any, Callable, Final, Literal

from zigpy.device import Device as ZigpyDevice
import zigpy.exceptions
from zigpy.zcl import Cluster as ZigpyCluster
from zigpy.zcl.foundation import Status

from zhaws.event import EventBase
from zhaws.model import BaseEvent
from zhaws.server.const import EVENT, EVENT_TYPE, EventTypes
from zhaws.server.util import LogMixin
from zhaws.server.zigbee.cluster.const import (
    CLUSTER_HANDLER_ZDO,
    CLUSTER_READS_PER_REQ,
    REPORT_CONFIG_ATTR_PER_REQ,
)
from zhaws.server.zigbee.cluster.decorators import decorate_command, retryable_request
from zhaws.server.zigbee.cluster.util import safe_read

if TYPE_CHECKING:
    from zhaws.server.zigbee.device import Device
    from zhaws.server.zigbee.endpoint import Endpoint

_LOGGER = logging.getLogger(__name__)

SIGNAL_ATTR_UPDATED: Final[str] = "attribute_updated"
CLUSTER_HANDLER_EVENT = "cluster_handler_event"
CLUSTER_HANDLER_ATTRIBUTE_UPDATED = "cluster_handler_attribute_updated"
CLUSTER_HANDLER_STATE_CHANGED = "cluster_handler_state_changed"


class ClusterHandlerStatus(Enum):
    """Status of a cluster handler."""

    CREATED = 1
    CONFIGURED = 2
    INITIALIZED = 3


class ClusterAttributeUpdatedEvent(BaseEvent):
    """Event to signal that a cluster attribute has been updated."""

    id: int
    name: str
    value: Any
    event_type: Literal["cluster_handler_event"] = "cluster_handler_event"
    event: Literal[
        "cluster_handler_attribute_updated"
    ] = "cluster_handler_attribute_updated"


class ClusterHandler(LogMixin, EventBase):
    """Base cluster handler for a Zigbee cluster handler."""

    REPORT_CONFIG: list[dict[int | str, str | tuple[int, int, int | float]]] = []
    BIND: bool = True

    # Dict of attributes to read on cluster handler initialization.
    # Dict keys -- attribute ID or names, with bool value indicating whether a cached
    # attribute read is acceptable.
    ZCL_INIT_ATTRS: dict[int | str, bool] = {}

    def __init__(self, cluster: ZigpyCluster, endpoint: Endpoint):
        """Initialize ClusterHandler."""
        super().__init__()
        self._generic_id: str = f"channel_0x{cluster.cluster_id:04x}"
        self._endpoint: Endpoint = endpoint
        self._cluster: ZigpyCluster = cluster
        self._id: str = f"{endpoint.id}:0x{cluster.cluster_id:04x}"
        unique_id: str = endpoint.unique_id.replace("-", ":")
        self._unique_id = f"{unique_id}:0x{cluster.cluster_id:04x}"
        if not hasattr(self, "_value_attribute") and self.REPORT_CONFIG:
            attr = self.REPORT_CONFIG[0].get("attr")
            if isinstance(attr, str):
                self.value_attribute = self.cluster.attridx.get(attr)
            else:
                self.value_attribute = attr
        self._status: ClusterHandlerStatus = ClusterHandlerStatus.CREATED
        self._cluster.add_listener(self)
        self.data_cache: dict[str, Any] = {}

    @property
    def id(self) -> str:
        """Return cluster handler id."""
        return self._id

    @property
    def generic_id(self) -> str:
        """Return the generic id for this cluster handler."""
        return self._generic_id

    @property
    def unique_id(self) -> str:
        """Return the unique id for this cluster handler."""
        return self._unique_id

    @property
    def cluster(self) -> ZigpyCluster:
        """Return the zigpy cluster for this cluster handler."""
        return self._cluster

    @property
    def name(self) -> str:
        """Return friendly name."""
        return self.cluster.ep_attribute or self._generic_id

    @property
    def status(self) -> ClusterHandlerStatus:
        """Return the status of the cluster handler."""
        return self._status

    def __hash__(self) -> int:
        """Make this a hashable."""
        return hash(self._unique_id)

    def send_event(self, signal: dict[str, Any]) -> None:
        """Broadcast an event from this cluster handler."""
        signal["cluster_handler"] = {
            "cluster": {
                "id": self.cluster.cluster_id,
                "name": self.cluster.name,
                "endpoint_attribute": self.cluster.ep_attribute,
                "endpoint_id": self.cluster.endpoint.endpoint_id,
            },
            "unique_id": self.unique_id,
        }
        self._endpoint.send_event(signal)

    async def bind(self) -> None:
        """Bind a zigbee cluster.

        This also swallows ZigbeeException exceptions that are thrown when
        devices are unreachable.
        """
        try:
            res = await self.cluster.bind()
            self.debug("bound '%s' cluster: %s", self.cluster.ep_attribute, res[0])
            """ TODO
            async_dispatcher_send(
                self._ch_pool.hass,
                ZHA_CHANNEL_MSG,
                {
                    ATTR_TYPE: ZHA_CHANNEL_MSG_BIND,
                    ZHA_CHANNEL_MSG_DATA: {
                        "cluster_name": self.cluster.name,
                        "cluster_id": self.cluster.cluster_id,
                        "success": res[0] == 0,
                    },
                },
            )
            """
        except (zigpy.exceptions.ZigbeeException, asyncio.TimeoutError) as ex:
            self.debug(
                "Failed to bind '%s' cluster: %s", self.cluster.ep_attribute, str(ex)
            )
            """ TODO
            async_dispatcher_send(
                self._ch_pool.hass,
                ZHA_CHANNEL_MSG,
                {
                    ATTR_TYPE: ZHA_CHANNEL_MSG_BIND,
                    ZHA_CHANNEL_MSG_DATA: {
                        "cluster_name": self.cluster.name,
                        "cluster_id": self.cluster.cluster_id,
                        "success": False,
                    },
                },
            )
            """

    async def configure_reporting(self) -> None:
        """Configure attribute reporting for a cluster.

        This also swallows ZigbeeException exceptions that are thrown when
        devices are unreachable.
        """
        event_data = {}
        kwargs = {}
        if (
            self.cluster.cluster_id >= 0xFC00
            and self._endpoint.device.manufacturer_code
        ):
            kwargs["manufacturer"] = self._endpoint.device.manufacturer_code

        for attr_report in self.REPORT_CONFIG:
            attr, config = attr_report["attr"], attr_report["config"]
            attr_name = self.cluster.attributes.get(attr, [attr])[0]
            event_data[attr_name] = {
                "min": config[0],
                "max": config[1],
                "id": attr,
                "name": attr_name,
                "change": config[2],
                "success": False,
            }

        to_configure = [*self.REPORT_CONFIG]
        chunk, rest = (
            to_configure[:REPORT_CONFIG_ATTR_PER_REQ],
            to_configure[REPORT_CONFIG_ATTR_PER_REQ:],
        )
        while chunk:
            reports = {rec["attr"]: rec["config"] for rec in chunk}
            try:
                res = await self.cluster.configure_reporting_multiple(reports, **kwargs)
                self._configure_reporting_status(reports, res[0])  # type: ignore
                # if we get a response, then it's a success
                for attr_stat in event_data.values():
                    attr_stat["success"] = True
            except (zigpy.exceptions.ZigbeeException, asyncio.TimeoutError) as ex:
                self.debug(
                    "failed to set reporting on '%s' cluster for: %s",
                    self.cluster.ep_attribute,
                    str(ex),
                )
                break
            chunk, rest = (
                rest[:REPORT_CONFIG_ATTR_PER_REQ],
                rest[REPORT_CONFIG_ATTR_PER_REQ:],
            )

        """ TODO
        async_dispatcher_send(
            self._ch_pool.hass,
            ZHA_CHANNEL_MSG,
            {
                ATTR_TYPE: ZHA_CHANNEL_MSG_CFG_RPT,
                ZHA_CHANNEL_MSG_DATA: {
                    "cluster_name": self.cluster.name,
                    "cluster_id": self.cluster.cluster_id,
                    "attributes": event_data,
                },
            },
        )
        """

    def _configure_reporting_status(
        self, attrs: dict[int | str, tuple], res: list | tuple
    ) -> None:
        """Parse configure reporting result."""
        if not isinstance(res, list):
            # assume default response
            self.debug(
                "attr reporting for '%s' on '%s': %s",
                attrs,
                self.name,
                res,
            )
            return
        if res[0].status == Status.SUCCESS and len(res) == 1:
            self.debug(
                "Successfully configured reporting for '%s' on '%s' cluster: %s",
                attrs,
                self.name,
                res,
            )
            return

        failed = [
            self.cluster.attributes.get(r.attrid, [r.attrid])[0]
            for r in res
            if r.status != Status.SUCCESS
        ]
        attrs = {self.cluster.attributes.get(r, [r])[0] for r in attrs}  # type: ignore
        self.debug(
            "Successfully configured reporting for '%s' on '%s' cluster",
            attrs - set(failed),  # type: ignore
            self.name,
        )
        self.debug(
            "Failed to configure reporting for '%s' on '%s' cluster: %s",
            failed,
            self.name,
            res,
        )

    async def async_configure(self) -> None:
        """Set cluster binding and attribute reporting."""
        if not self._endpoint.device.skip_configuration:
            if self.BIND:
                await self.bind()
            if self.cluster.is_server:
                await self.configure_reporting()
            ch_specific_cfg = getattr(self, "async_configure_handler_specific", None)
            if ch_specific_cfg:
                await ch_specific_cfg()
            self.debug("finished cluster handler configuration")
        else:
            self.debug("skipping cluster handler configuration")
        self._status = ClusterHandlerStatus.CONFIGURED

    @retryable_request(delays=(1, 1, 3))
    async def async_initialize(self, from_cache: bool) -> None:
        """Initialize cluster handler."""
        if not from_cache and self._endpoint.device.skip_configuration:
            self._status = ClusterHandlerStatus.INITIALIZED
            return

        self.debug("initializing cluster handler: from_cache: %s", from_cache)
        cached = [a for a, cached in self.ZCL_INIT_ATTRS.items() if cached]
        uncached = [a for a, cached in self.ZCL_INIT_ATTRS.items() if not cached]
        uncached.extend([cfg["attr"] for cfg in self.REPORT_CONFIG])  # type: ignore #TODO see if this can be fixed

        if cached:
            await self._get_attributes(True, cached, from_cache=True)
        if uncached:
            await self._get_attributes(True, uncached, from_cache=from_cache)

        ch_specific_init = getattr(self, "async_initialize_handler_specific", None)
        if ch_specific_init:
            await ch_specific_init(from_cache=from_cache)

        self.debug("finished cluster handler initialization")
        self._status = ClusterHandlerStatus.INITIALIZED

    def cluster_command(self, tsn: int, command_id: int, args: Any) -> None:
        """Handle commands received to this cluster."""
        _LOGGER.info("received command %s args %s", command_id, args)

    def attribute_updated(self, attrid: int, value: Any) -> None:
        """Handle attribute updates on this cluster."""
        self.send_event(
            {
                EVENT: SIGNAL_ATTR_UPDATED,
                EVENT_TYPE: EventTypes.RAW_ZCL_EVENT,
                "attribute": {
                    "id": attrid,
                    "name": self.cluster.attributes.get(attrid, [attrid])[0],
                    "value": value,
                },
            }
        )

        self.emit(
            CLUSTER_HANDLER_EVENT,
            ClusterAttributeUpdatedEvent(
                id=attrid,
                name=self.cluster.attributes.get(attrid, [attrid])[0],
                value=value,
            ),
        )

    def zdo_command(self, *args: Any, **kwargs: Any) -> None:
        """Handle ZDO commands on this cluster."""

    def zha_send_event(self, command: str, args: int | dict) -> None:
        """Relay events to hass."""
        """ TODO
        self._ch_pool.zha_send_event(
            {
                ATTR_UNIQUE_ID: self.unique_id,
                ATTR_CLUSTER_ID: self.cluster.cluster_id,
                ATTR_COMMAND: command,
                ATTR_ARGS: args,
            }
        )
        """

    async def async_update(self) -> None:
        """Retrieve latest state from cluster."""

    async def get_attribute_value(
        self, attribute: int | str, from_cache: bool = True
    ) -> Any:
        """Get the value for an attribute."""
        manufacturer = None
        manufacturer_code = self._endpoint.device.manufacturer_code
        if self.cluster.cluster_id >= 0xFC00 and manufacturer_code:
            manufacturer = manufacturer_code
        result = await safe_read(
            self._cluster,
            [attribute],
            allow_cache=from_cache,
            only_cache=from_cache and not self._endpoint.device.is_mains_powered,
            manufacturer=manufacturer,
        )
        return result.get(attribute)

    async def _get_attributes(
        self,
        raise_exceptions: bool,
        attributes: list[int | str],
        from_cache: bool = True,
    ) -> dict[int | str, Any]:
        """Get the values for a list of attributes."""
        manufacturer = None
        manufacturer_code = self._endpoint.device.manufacturer_code
        if self.cluster.cluster_id >= 0xFC00 and manufacturer_code:
            manufacturer = manufacturer_code
        chunk = attributes[:CLUSTER_READS_PER_REQ]
        rest = attributes[CLUSTER_READS_PER_REQ:]
        result = {}
        while chunk:
            try:
                read, _ = await self.cluster.read_attributes(
                    attributes,
                    allow_cache=from_cache,
                    only_cache=from_cache
                    and not self._endpoint.device.is_mains_powered,
                    manufacturer=manufacturer,
                )
                result.update(read)
            except (asyncio.TimeoutError, zigpy.exceptions.ZigbeeException) as ex:
                self.debug(
                    "failed to get attributes '%s' on '%s' cluster: %s",
                    attributes,
                    self.cluster.ep_attribute,
                    str(ex),
                )
                if raise_exceptions:
                    raise
            chunk = rest[:CLUSTER_READS_PER_REQ]
            rest = rest[CLUSTER_READS_PER_REQ:]
        return result

    get_attributes = partialmethod(_get_attributes, False)

    def to_json(self) -> dict:
        """Return JSON representation of this cluster handler."""
        json = {
            "class_name": self.__class__.__name__,
            "generic_id": self._generic_id,
            "endpoint_id": self._endpoint.id,
            "cluster": {
                "id": self._cluster.cluster_id,
                "name": self._cluster.name,
                "type": "client" if self._cluster.is_client else "server",
                "commands": self._cluster.commands,
            },
            "id": self._id,
            "unique_id": self._unique_id,
            "status": self._status.name,
        }

        if hasattr(self, "value_attribute"):
            json["value_attribute"] = self.value_attribute

        return json

    def log(self, level: int, msg: str, *args: Any, **kwargs: Any) -> None:
        """Log a message."""
        msg = f"[%s:%s]: {msg}"
        args = (self._endpoint.device.nwk, self._id) + args
        _LOGGER.log(level, msg, *args, **kwargs)

    def __getattr__(self, name: str) -> Any:
        """Get attribute or a decorated cluster command."""
        if hasattr(self._cluster, name) and callable(getattr(self._cluster, name)):
            command: Callable = getattr(self._cluster, name)
            command.__name__ = name
            return decorate_command(self, command)
        return self.__getattribute__(name)


class ZDOClusterHandler(LogMixin):
    """Cluster handler for ZDO events."""

    def __init__(self, device: Device):
        """Initialize ZDOClusterHandler."""
        self.name: str = CLUSTER_HANDLER_ZDO
        self._cluster: ZigpyCluster = device.device.endpoints[0]
        self._device: Device = device
        self._status: ClusterHandlerStatus = ClusterHandlerStatus.CREATED
        self._unique_id: str = f"{str(device.ieee)}:{device.name}_ZDO"
        self._cluster.add_listener(self)

    @property
    def unique_id(self) -> str:
        """Return the unique id for this cluster handler."""
        return self._unique_id

    @property
    def cluster(self) -> ZigpyCluster:
        """Return the aigpy cluster for this cluster handler."""
        return self._cluster

    @property
    def status(self) -> ClusterHandlerStatus:
        """Return the status of the cluster handler."""
        return self._status

    def device_announce(self, zigpy_device: ZigpyDevice) -> None:
        """Device announce handler."""

    def permit_duration(self, duration: int) -> None:
        """Permit handler."""

    async def async_initialize(self, from_cache: bool) -> None:
        """Initialize cluster handler."""
        self._status = ClusterHandlerStatus.INITIALIZED

    async def async_configure(self) -> None:
        """Configure cluster handler."""
        self._status = ClusterHandlerStatus.CONFIGURED

    def log(self, level: int, msg: str, *args: Any, **kwargs: Any) -> None:
        """Log a message."""
        msg = f"[%s:ZDO](%s): {msg}"
        args = (self._device.nwk, self._device.model) + args
        _LOGGER.log(level, msg, *args, **kwargs)


class ClientClusterHandler(ClusterHandler):
    """Cluster handler for Zigbee client (output) clusters."""

    def attribute_updated(self, attrid: int, value: Any) -> None:
        """Handle an attribute updated on this cluster."""
        """ TODO
        self.zha_send_event(
            SIGNAL_ATTR_UPDATED,
            {
                ATTR_ATTRIBUTE_ID: attrid,
                ATTR_ATTRIBUTE_NAME: self._cluster.attributes.get(attrid, ["Unknown"])[
                    0
                ],
                ATTR_VALUE: value,
            },
        )
        """

    def cluster_command(self, tsn: int, command_id: int, args: Any) -> None:
        """Handle a cluster command received on this cluster."""
        if (
            self._cluster.server_commands is not None
            and self._cluster.server_commands.get(command_id) is not None
        ):
            self.zha_send_event(self._cluster.server_commands.get(command_id)[0], args)



"""Controller implementation for the zhaws.client."""

from __future__ import annotations

import logging
from types import TracebackType
from typing import Any

from aiohttp import ClientSession
from async_timeout import timeout

from zhaws.client.client import Client
from zhaws.client.device import Device
from zhaws.client.group import Group
from zhaws.client.helpers import (
    AlarmControlPanelHelper,
    ButtonHelper,
    ClientHelper,
    ClimateHelper,
    CoverHelper,
    DeviceHelper,
    FanHelper,
    GroupHelper,
    LightHelper,
    LockHelper,
    NetworkHelper,
    NumberHelper,
    PlatformEntityHelper,
    SelectHelper,
    ServerHelper,
    SirenHelper,
    SwitchHelper,
)
from zhaws.client.model.commands import Command, CommandResponse
from zhaws.client.model.events import (
    DeviceConfiguredEvent,
    DeviceFullyInitializedEvent,
    DeviceJoinedEvent,
    DeviceLeftEvent,
    DeviceRemovedEvent,
    PlatformEntityEvent,
    RawDeviceInitializedEvent,
)
from zhaws.event import EventBase

CONNECT_TIMEOUT = 10

_LOGGER = logging.getLogger(__name__)


class Controller(EventBase):
    """Controller implementation."""

    def __init__(
        self, ws_server_url: str, aiohttp_session: ClientSession | None = None
    ):
        super().__init__()
        self._ws_server_url: str = ws_server_url
        self._client: Client = Client(ws_server_url, aiohttp_session)
        self._devices: dict[str, Device] = {}
        self._groups: dict[int, Group] = {}
        self._client.on_event(
            "platform_entity_event", self.handle_platform_entity_event
        )
        self._client.on_event("controller_event", self._handle_event_protocol)
        self.lights: LightHelper = LightHelper(self._client)
        self.switches: SwitchHelper = SwitchHelper(self._client)
        self.sirens: SirenHelper = SirenHelper(self._client)
        self.buttons: ButtonHelper = ButtonHelper(self._client)
        self.covers: CoverHelper = CoverHelper(self._client)
        self.fans: FanHelper = FanHelper(self._client)
        self.locks: LockHelper = LockHelper(self._client)
        self.numbers: NumberHelper = NumberHelper(self._client)
        self.selects: SelectHelper = SelectHelper(self._client)
        self.thermostats: ClimateHelper = ClimateHelper(self._client)
        self.alarm_control_panels: AlarmControlPanelHelper = AlarmControlPanelHelper(
            self._client
        )
        self.entities: PlatformEntityHelper = PlatformEntityHelper(self._client)
        self.clients: ClientHelper = ClientHelper(self._client)
        self.groups_helper: GroupHelper = GroupHelper(self._client)
        self.devices_helper: DeviceHelper = DeviceHelper(self._client)
        self.network: NetworkHelper = NetworkHelper(self._client)
        self.server_helper: ServerHelper = ServerHelper(self._client)

    @property
    def devices(self) -> dict[str, Device]:
        """Return the devices."""
        return self._devices

    @property
    def groups(self) -> dict[int, Group]:
        """Return the groups."""
        return self._groups

    async def connect(self) -> None:
        """Connect to the websocket server."""
        try:
            async with timeout(CONNECT_TIMEOUT):
                await self._client.connect()
        except Exception as err:
            _LOGGER.error("Unable to connect to the ZHA wss: %s", err)
            raise err

        await self._client.listen()

    async def disconnect(self) -> None:
        """Disconnect from the websocket server."""
        await self._client.disconnect()

    async def __aenter__(self) -> Controller:
        """Connect to the websocket server."""
        await self.connect()
        return self

    async def __aexit__(
        self, exc_type: Exception, exc_value: str, traceback: TracebackType
    ) -> None:
        """Disconnect from the websocket server."""
        await self.disconnect()

    async def send_command(self, command: Command) -> CommandResponse:
        """Send a command and get a response."""
        return await self._client.async_send_command(command.dict(exclude_none=True))

    async def load_devices(self) -> None:
        """Load devices from the websocket server."""
        response_devices = await self.devices_helper.get_devices()
        for ieee, device in response_devices.items():
            self._devices[ieee] = Device(device, self, self._client)

    async def load_groups(self) -> None:
        """Load groups from the websocket server."""
        response_groups = await self.groups_helper.get_groups()
        for id, group in response_groups.items():
            self._groups[id] = Group(group, self, self._client)

    def handle_platform_entity_event(self, event: PlatformEntityEvent) -> None:
        """Handle a platform_entity_event from the websocket server."""
        _LOGGER.debug("platform_entity_event: %s", event)
        if event.device:
            device = self.devices.get(event.device.ieee)
            if device is None:
                _LOGGER.warning("Received event from unknown device: %s", event)
                return
            entity: Any = device.device.entities.get(event.platform_entity.unique_id)
        elif event.group:
            group = self.groups.get(event.group.id)
            if not group:
                _LOGGER.warning("Received event from unknown group: %s", event)
                return
            entity = group.group.entities.get(event.platform_entity.unique_id)
        if entity is None:
            _LOGGER.warning("Received event for an unknown entity: %s", event)
            return
        entity.state = event.state  # update our state locally and emit an event
        entity.emit(event.event, event)

    def handle_device_joined(self, event: DeviceJoinedEvent) -> None:
        """Handle device joined.

        At this point, no information about the device is known other than its
        address
        """
        _LOGGER.info("Device %s - %s joined", event.ieee, event.nwk)
        self.emit("device_joined", event)

    def handle_raw_device_initialized(self, event: RawDeviceInitializedEvent) -> None:
        """Handle a device initialization without quirks loaded."""
        _LOGGER.info("Device %s - %s raw device initialized", event.ieee, event.nwk)
        self.emit("raw_device_initialized", event)

    def handle_device_configured(self, event: DeviceConfiguredEvent) -> None:
        """Handle device configured event."""
        device = event.device
        _LOGGER.info("Device %s - %s configured", device.ieee, device.nwk)
        self.emit("device_configured", event)

    def handle_device_fully_initialized(
        self, event: DeviceFullyInitializedEvent
    ) -> None:
        """Handle device joined and basic information discovered."""
        device = event.device
        _LOGGER.info("Device %s - %s initialized", device.ieee, device.nwk)
        # TODO devise a way to update existing devices so we don't lose event subscriptions
        self._devices[device.ieee] = Device(device, self, self._client)
        self.emit("device_fully_initialized", event)

    def handle_device_left(self, event: DeviceLeftEvent) -> None:
        """Handle device leaving the network."""
        _LOGGER.info("Device %s - %s left", event.ieee, event.nwk)
        self.emit("device_left", event)

    def handle_device_removed(self, event: DeviceRemovedEvent) -> None:
        """Handle device being removed from the network."""
        device = event.device
        _LOGGER.info(
            "Device %s - %s has been removed from the network", device.ieee, device.nwk
        )
        self._devices.pop(device.ieee, None)
        self.emit("device_removed", event)

    def handle_group_member_removed(self, event: Any) -> None:
        """Handle group member removed event."""
        self.emit("group_member_removed", event)

    def handle_group_member_added(self, event: Any) -> None:
        """Handle group member added event."""
        self.emit("group_member_added", event)

    def handle_group_added(self, event: Any) -> None:
        """Handle group added event."""
        self.emit("group_added", event)

    def handle_group_removed(self, event: Any) -> None:
        """Handle group removed event."""
        self.emit("group_removed", event)



"""Platform module for zhawss."""
from __future__ import annotations

import abc
import asyncio
import logging
from typing import TYPE_CHECKING, Any, Type, Union

from zhaws.event import EventBase
from zhaws.server.const import EVENT, EVENT_TYPE, EventTypes, PlatformEntityEvents
from zhaws.server.platforms.model import STATE_CHANGED, EntityStateChangedEvent
from zhaws.server.platforms.registries import Platform

if TYPE_CHECKING:
    from zhaws.server.zigbee.cluster import ClusterHandler
    from zhaws.server.zigbee.device import Device
    from zhaws.server.zigbee.group import Group
    from zhaws.server.zigbee.endpoint import Endpoint

from zhaws.server.util import LogMixin

_LOGGER = logging.getLogger(__name__)


class BaseEntity(LogMixin, EventBase):
    """Base class for entities."""

    PLATFORM: Platform = Platform.UNKNOWN

    def __init__(
        self,
        unique_id: str,
    ):
        """Initialize the platform entity."""
        super().__init__()
        self._unique_id: str = unique_id
        self._previous_state: Any = None

    @property
    def unique_id(self) -> str:
        """Return the unique id."""
        return self._unique_id

    @abc.abstractmethod
    def send_event(self, signal: dict[str, Any]) -> None:
        """Broadcast an event from this platform entity."""

    @abc.abstractmethod
    def get_identifiers(self) -> dict[str, str | int]:
        """Return a dict with the information necessary to identify this entity."""

    def get_state(self) -> dict:
        """Return the arguments to use in the command."""
        return {
            "class_name": self.__class__.__name__,
        }

    async def async_update(self) -> None:
        """Retrieve latest state."""

    def maybe_send_state_changed_event(self) -> None:
        """Send the state of this platform entity."""
        state = self.get_state()
        if self._previous_state != state:
            self.send_event(
                {
                    "state": self.get_state(),
                    EVENT: PlatformEntityEvents.PLATFORM_ENTITY_STATE_CHANGED,
                    EVENT_TYPE: EventTypes.PLATFORM_ENTITY_EVENT,
                }
            )
            self.emit(
                STATE_CHANGED, EntityStateChangedEvent.parse_obj(self.get_identifiers())
            )
            self._previous_state = state

    def to_json(self) -> dict:
        """Return a JSON representation of the platform entity."""
        return {
            "unique_id": self._unique_id,
            "platform": self.PLATFORM,
            "class_name": self.__class__.__name__,
            "state": self.get_state(),
        }

    def log(self, level: int, msg: str, *args: Any, **kwargs: Any) -> None:
        """Log a message."""
        msg = f"%s: {msg}"
        args = (self._unique_id,) + args
        _LOGGER.log(level, msg, *args, **kwargs)


class PlatformEntity(BaseEntity):
    """Class that represents an entity for a device platform."""

    unique_id_suffix: Union[str, None] = None

    def __init_subclass__(
        cls: Type[PlatformEntity], id_suffix: Union[str, None] = None, **kwargs: Any
    ):
        """Initialize subclass.
        :param id_suffix: suffix to add to the unique_id of the entity. Used for multi
                          entities using the same cluster handler/cluster id for the entity.
        """
        super().__init_subclass__(**kwargs)
        if id_suffix:
            cls.unique_id_suffix = id_suffix

    def __init__(
        self,
        unique_id: str,
        cluster_handlers: list[ClusterHandler],
        endpoint: Endpoint,
        device: Device,
    ):
        """Initialize the platform entity."""
        super().__init__(unique_id)
        ieeetail = "".join([f"{o:02x}" for o in device.ieee[:4]])
        ch_names = ", ".join(sorted(ch.name for ch in cluster_handlers))
        self._name: str = f"{device.name} {ieeetail} {ch_names}"
        if self.unique_id_suffix:
            self._name += f" {self.unique_id_suffix}"
            self._unique_id += f"-{self.unique_id_suffix}"
        self._cluster_handlers: list[ClusterHandler] = cluster_handlers
        self.cluster_handlers: dict[str, ClusterHandler] = {}
        for cluster_handler in cluster_handlers:
            self.cluster_handlers[cluster_handler.name] = cluster_handler
        self._device: Device = device
        self._endpoint = endpoint
        self._device.platform_entities[self.unique_id] = self

    @classmethod
    def create_platform_entity(
        cls: Type[PlatformEntity],
        unique_id: str,
        cluster_handlers: list[ClusterHandler],
        endpoint: Endpoint,
        device: Device,
        **kwargs: Any,
    ) -> Union[PlatformEntity, None]:
        """Entity Factory.
        Return a platform entity if it is a supported configuration, otherwise return None
        """
        return cls(unique_id, cluster_handlers, endpoint, device, **kwargs)

    @property
    def device(self) -> Device:
        """Return the device."""
        return self._device

    @property
    def endpoint(self) -> Endpoint:
        """Return the endpoint."""
        return self._endpoint

    @property
    def should_poll(self) -> bool:
        """Return True if we need to poll for state changes."""
        return False

    @property
    def available(self) -> bool:
        """Return true if the device this entity belongs to is available."""
        return self.device.available

    @property
    def name(self) -> str:
        """Return the name of the platform entity."""
        return self._name

    def get_identifiers(self) -> dict[str, str | int]:
        """Return a dict with the information necessary to identify this entity."""
        return {
            "unique_id": self.unique_id,
            "platform": self.PLATFORM,
            "device_ieee": str(self.device.ieee),
            "endpoint_id": self.endpoint.id,
        }

    def send_event(self, signal: dict[str, Any]) -> None:
        """Broadcast an event from this platform entity."""
        signal["platform_entity"] = {
            "name": self._name,
            "unique_id": self._unique_id,
            "platform": self.PLATFORM,
        }
        signal["endpoint"] = {
            "id": self._endpoint.id,
        }
        _LOGGER.info("Sending event from platform entity: %s", signal)
        self.device.send_event(signal)

    def to_json(self) -> dict:
        """Return a JSON representation of the platform entity."""
        json = super().to_json()
        json["name"] = self._name
        json["cluster_handlers"] = [ch.to_json() for ch in self._cluster_handlers]
        json["device_ieee"] = str(self._device.ieee)
        json["endpoint_id"] = self._endpoint.id
        return json

    async def async_update(self) -> None:
        """Retrieve latest state."""
        tasks = [
            cluster_handler.async_update()
            for cluster_handler in self.cluster_handlers.values()
            if hasattr(cluster_handler, "async_update")
        ]
        if tasks:
            await asyncio.gather(*tasks)
            self.maybe_send_state_changed_event()


class GroupEntity(BaseEntity):
    """A base class for group entities."""

    def __init__(
        self,
        group: Group,
    ) -> None:
        """Initialize a group."""
        super().__init__(f"{self.PLATFORM}.{group.group_id}")
        self._zigpy_group: Group = group
        self._name: str = f"{group.name}_0x{group.group_id:04x}"
        self._group: Group = group
        self._group.register_group_entity(self)

    @property
    def name(self) -> str:
        """Return the name of the group entity."""
        return self._name

    @property
    def group_id(self) -> int:
        """Return the group id."""
        return self._group.group_id

    @property
    def group(self) -> Group:
        """Return the group."""
        return self._group

    def get_identifiers(self) -> dict[str, str | int]:
        """Return a dict with the information necessary to identify this entity."""
        return {
            "unique_id": self.unique_id,
            "platform": self.PLATFORM,
            "group_id": self.group.group_id,
        }

    def update(self, _: Any | None = None) -> None:
        """Update the state of this group entity."""

    def send_event(self, signal: dict[str, Any]) -> None:
        """Broadcast an event from this group entity."""
        signal["platform_entity"] = {
            "name": self._name,
            "unique_id": self._unique_id,
            "platform": self.PLATFORM,
        }
        signal["group"] = {
            "id": self._group.group_id,
        }
        _LOGGER.info("Sending event from group entity: %s", signal)
        self._group.send_event(signal)

    def to_json(self) -> dict[str, Any]:
        """Return a JSON representation of the group."""
        json = super().to_json()
        json["name"] = self._name
        json["group_id"] = self.group_id
        return json



"""Select platform for zhawss."""
from __future__ import annotations

from enum import Enum
import functools
from typing import TYPE_CHECKING, Any, Final, Union

import zigpy.types as t
from zigpy.zcl.clusters.security import IasWd

from zhaws.server.platforms import PlatformEntity
from zhaws.server.platforms.registries import PLATFORM_ENTITIES, Platform
from zhaws.server.zigbee.cluster.const import CLUSTER_HANDLER_IAS_WD

if TYPE_CHECKING:
    from zhaws.server.zigbee.cluster import ClusterHandler
    from zhaws.server.zigbee.device import Device
    from zhaws.server.zigbee.endpoint import Endpoint

MULTI_MATCH = functools.partial(PLATFORM_ENTITIES.multipass_match, Platform.SELECT)
STATE_UNKNOWN: Final[str] = "unknown"


class Strobe(t.enum8):  # type: ignore #TODO fix type
    No_Strobe = 0x00
    Strobe = 0x01


class EnumSelect(PlatformEntity):

    PLATFORM = Platform.SELECT
    _enum: type[Enum]

    def __init__(
        self,
        unique_id: str,
        cluster_handlers: list[ClusterHandler],
        endpoint: Endpoint,
        device: Device,
    ):
        """Initialize the select entity."""
        self._attr_name = self._enum.__name__
        self._attr_options = [entry.name.replace("_", " ") for entry in self._enum]
        super().__init__(unique_id, cluster_handlers, endpoint, device)
        self._cluster_handler: ClusterHandler = cluster_handlers[0]

        """TODO
        self._cluster_handler.data_cache[self._attr_name] = self._enum[
                last_state.state.replace(" ", "_")
            ]
        """

    @property
    def current_option(self) -> Union[str, None]:
        """Return the selected entity option to represent the entity state."""
        option = self._cluster_handler.data_cache.get(self._attr_name)
        if option is None:
            return None
        return option.name.replace("_", " ")

    async def async_select_option(self, option: Union[str, int], **kwargs: Any) -> None:
        """Change the selected option."""
        if isinstance(option, str):
            self._cluster_handler.data_cache[self._attr_name] = self._enum[
                option.replace(" ", "_")
            ]
            self.maybe_send_state_changed_event()

    def to_json(self) -> dict:
        """Return a JSON representation of the select."""
        json = super().to_json()
        json["enum"] = self._enum.__name__
        json["options"] = self._attr_options
        return json

    def get_state(self) -> dict:
        """Return the state of the select."""
        response = super().get_state()
        response["state"] = self.current_option
        return response


@MULTI_MATCH(cluster_handler_names=CLUSTER_HANDLER_IAS_WD)
class DefaultToneSelectEntity(EnumSelect, id_suffix=IasWd.Warning.WarningMode.__name__):
    """Representation of a zhawss default siren tone select entity."""

    _enum: IasWd.Warning.WarningMode = IasWd.Warning.WarningMode


@MULTI_MATCH(cluster_handler_names=CLUSTER_HANDLER_IAS_WD)
class DefaultSirenLevelSelectEntity(
    EnumSelect, id_suffix=IasWd.Warning.SirenLevel.__name__
):
    """Representation of a zhawss default siren level select entity."""

    _enum: IasWd.Warning.SirenLevel = IasWd.Warning.SirenLevel


@MULTI_MATCH(cluster_handler_names=CLUSTER_HANDLER_IAS_WD)
class DefaultStrobeLevelSelectEntity(EnumSelect, id_suffix=IasWd.StrobeLevel.__name__):
    """Representation of a zhawss default siren strobe level select entity."""

    _enum = IasWd.StrobeLevel


@MULTI_MATCH(cluster_handler_names=CLUSTER_HANDLER_IAS_WD)
class DefaultStrobeSelectEntity(EnumSelect, id_suffix=Strobe.__name__):
    """Representation of a zhawss default siren strobe select entity."""

    _enum = Strobe



"""Client classes for zhawss."""
from __future__ import annotations

import asyncio
import json
import logging
from typing import TYPE_CHECKING, Any, Callable, Optional, Tuple, Union

import voluptuous
from websockets.server import WebSocketServerProtocol

from zhaws.backports.enum import StrEnum
from zhaws.server.const import (
    COMMAND,
    ERROR_CODE,
    ERROR_MESSAGE,
    EVENT_TYPE,
    MESSAGE_ID,
    MESSAGE_TYPE,
    MINIMAL_MESSAGE_SCHEMA,
    SUCCESS,
    WEBSOCKET_API,
    ZIGBEE_ERROR_CODE,
    EventTypes,
    MessageTypes,
)
from zhaws.server.websocket.api import decorators, register_api_command

if TYPE_CHECKING:
    from zhaws.server.websocket.server import Server

_LOGGER = logging.getLogger(__name__)


class ClientAPICommands(StrEnum):
    """Enum for all client API commands."""

    LISTEN_RAW_ZCL = "client_listen_raw_zcl"
    LISTEN = "client_listen"
    DISCONNECT = "client_disconnect"


class Client:
    """ZHAWSS client implementation."""

    def __init__(
        self,
        websocket: WebSocketServerProtocol,
        client_manager: ClientManager,
    ):
        """Initialize the client."""
        self._websocket: WebSocketServerProtocol = websocket
        self._client_manager: ClientManager = client_manager
        self.receive_events: bool = False
        self.receive_raw_zcl_events: bool = False

    @property
    def is_connected(self) -> bool:
        """Return True if the websocket connection is connected."""
        return self._websocket.open

    def disconnect(self) -> None:
        """Disconnect this client and close the websocket."""
        asyncio.create_task(self._websocket.close())

    def send_event(self, message: dict[str, Any]) -> None:
        """Send event data to this client."""
        message[MESSAGE_TYPE] = MessageTypes.EVENT
        self._send_data(message)

    def send_result_success(
        self, request_message: dict[str, Any], data: Optional[dict[str, Any]] = None
    ) -> None:
        """Send success result prompted by a client request."""
        message = {
            SUCCESS: True,
            MESSAGE_ID: request_message[MESSAGE_ID],
            MESSAGE_TYPE: MessageTypes.RESULT,
            COMMAND: request_message[COMMAND],
        }
        if data:
            message.update(data)
        self._send_data(message)

    def send_result_error(
        self, request_message: dict[str, Any], error_code: str, error_message: str
    ) -> None:
        """Send error result prompted by a client request."""
        message = {
            SUCCESS: False,
            MESSAGE_ID: request_message[MESSAGE_ID],
            MESSAGE_TYPE: MessageTypes.RESULT,
            COMMAND: request_message[COMMAND],
            ERROR_CODE: error_code,
            ERROR_MESSAGE: error_message,
        }
        self._send_data(message)

    def send_result_zigbee_error(
        self,
        request_message: dict[str, Any],
        error_message: str,
        zigbee_error_code: str,
    ) -> None:
        """Send zigbee error result prompted by a client zigbee request."""
        self.send_result_error(
            request_message={
                **request_message,
                ZIGBEE_ERROR_CODE: zigbee_error_code,
            },
            error_code="zigbee_error",
            error_message=error_message,
        )

    def _send_data(self, data: dict[str, Any]) -> None:
        """Send data to this client."""
        try:
            message = json.dumps(data)
        except TypeError as exc:
            _LOGGER.error("Couldn't serialize data: %s", data, exc_info=exc)
        else:
            asyncio.create_task(self._websocket.send(message))

    async def _handle_incoming_message(self, message: Union[str, bytes]) -> None:
        """Handle an incoming message."""
        _LOGGER.info("Message received: %s", message)
        handlers: dict[
            str, Tuple[Callable, Callable]
        ] = self._client_manager.server.data[WEBSOCKET_API]

        loaded_message = json.loads(message)
        _LOGGER.debug(
            "Received message: %s on websocket: %s", loaded_message, self._websocket.id
        )

        try:
            msg = MINIMAL_MESSAGE_SCHEMA(loaded_message)
        except voluptuous.Invalid as exception:
            _LOGGER.error(
                f"Received invalid command[unable to parse command]: {loaded_message}",
                exc_info=exception,
            )
            return

        if msg[COMMAND] not in handlers:
            _LOGGER.error(
                f"Received invalid command[command not registered]: {msg[COMMAND]}"
            )
            return

        handler, schema = handlers[msg[COMMAND]]

        try:
            handler(self._client_manager.server, self, schema(msg))
        except Exception as err:  # pylint: disable=broad-except
            # TODO Fix this - make real error codes with error messages
            _LOGGER.error("Error handling message: %s", msg, exc_info=err)
            self.send_result_error(
                loaded_message, "INTERNAL_ERROR", f"Internal error: {err}"
            )

    async def listen(self) -> None:
        async for message in self._websocket:
            asyncio.create_task(self._handle_incoming_message(message))

    def will_accept_message(self, message: dict[str, Any]) -> bool:
        """Checks if the client has registered to accept this type of message."""
        if not self.receive_events:
            return False

        if (
            message[EVENT_TYPE] == EventTypes.RAW_ZCL_EVENT
            and not self.receive_raw_zcl_events
        ):
            _LOGGER.info(
                "Client %s not accepting raw ZCL events: %s",
                self._websocket.id,
                message,
            )
            return False

        return True


@decorators.websocket_command(
    {
        COMMAND: str(ClientAPICommands.LISTEN_RAW_ZCL),
    }
)
@decorators.async_response
async def listen_raw_zcl(
    server: Server, client: Client, message: dict[str, Any]
) -> None:
    """Listen for raw ZCL events."""
    client.receive_raw_zcl_events = True
    client.send_result_success(message)


@decorators.websocket_command(
    {
        COMMAND: str(ClientAPICommands.LISTEN),
    }
)
@decorators.async_response
async def listen(server: Server, client: Client, message: dict[str, Any]) -> None:
    """Listen for events."""
    client.receive_events = True
    client.send_result_success(message)


@decorators.websocket_command(
    {
        COMMAND: str(ClientAPICommands.DISCONNECT),
    }
)
@decorators.async_response
async def disconnect(server: Server, client: Client, message: dict[str, Any]) -> None:
    """Disconnect the client."""
    client.disconnect()
    server.client_manager.remove_client(client)


def load_api(server: Server) -> None:
    """Load the api command handlers."""
    register_api_command(server, listen_raw_zcl)
    register_api_command(server, listen)
    register_api_command(server, disconnect)


class ClientManager:
    """ZHAWSS client manager implementation."""

    def __init__(self, server: Server):
        """Initialize the client."""
        self._server: Server = server
        self._clients: list[Client] = []

    @property
    def server(self) -> Server:
        """Return the server this ClientManager belongs to."""
        return self._server

    async def add_client(self, websocket: WebSocketServerProtocol) -> None:
        """Adds a new client to the client manager."""
        client: Client = Client(websocket, self)
        self._clients.append(client)
        await client.listen()

    def remove_client(self, client: Client) -> None:
        """Removes a client from the client manager."""
        client.disconnect()
        self._clients.remove(client)

    def broadcast(self, message: dict[str, Any]) -> None:
        """Broadcast a message to all connected clients."""
        clients_to_remove = []

        for client in self._clients:
            if not client.is_connected:
                # XXX: We cannot remove elements from `_clients` while iterating over it
                clients_to_remove.append(client)
                continue

            if not client.will_accept_message(message):
                continue

            _LOGGER.info(
                "Broadcasting message: %s to client: %s",
                message,
                client._websocket.id,
            )
            """TODO use the receive flags on the client to determine if the client should receive the message"""
            client.send_event(message)

        for client in clients_to_remove:
            self.remove_client(client)



"""Climate platform for zhawss."""
from __future__ import annotations

import asyncio
from datetime import datetime
import functools
import logging
from typing import TYPE_CHECKING, Any, Final, Optional, Union

from zigpy.zcl.clusters.hvac import Fan as FanCluster, Thermostat as ThermostatCluster

from zhaws.backports.enum import StrEnum
from zhaws.server.decorators import periodic
from zhaws.server.platforms import PlatformEntity
from zhaws.server.platforms.registries import PLATFORM_ENTITIES, Platform
from zhaws.server.zigbee.cluster import (
    CLUSTER_HANDLER_EVENT,
    ClusterAttributeUpdatedEvent,
)
from zhaws.server.zigbee.cluster.const import (
    CLUSTER_HANDLER_FAN,
    CLUSTER_HANDLER_THERMOSTAT,
)

if TYPE_CHECKING:
    from zhaws.server.zigbee.cluster import ClusterHandler
    from zhaws.server.zigbee.device import Device
    from zhaws.server.zigbee.endpoint import Endpoint

STRICT_MATCH = functools.partial(PLATFORM_ENTITIES.strict_match, Platform.CLIMATE)
MULTI_MATCH = functools.partial(PLATFORM_ENTITIES.multipass_match, Platform.CLIMATE)


ATTR_SYS_MODE: Final[str] = "system_mode"
ATTR_FAN_MODE: Final[str] = "fan_mode"
ATTR_RUNNING_MODE: Final[str] = "running_mode"
ATTR_SETPT_CHANGE_SRC: Final[str] = "setpoint_change_source"
ATTR_SETPT_CHANGE_AMT: Final[str] = "setpoint_change_amount"
ATTR_OCCUPANCY: Final[str] = "occupancy"
ATTR_PI_COOLING_DEMAND: Final[str] = "pi_cooling_demand"
ATTR_PI_HEATING_DEMAND: Final[str] = "pi_heating_demand"
ATTR_OCCP_COOL_SETPT: Final[str] = "occupied_cooling_setpoint"
ATTR_OCCP_HEAT_SETPT: Final[str] = "occupied_heating_setpoint"
ATTR_UNOCCP_HEAT_SETPT: Final[str] = "unoccupied_heating_setpoint"
ATTR_UNOCCP_COOL_SETPT: Final[str] = "unoccupied_cooling_setpoint"
ATTR_HVAC_MODE: Final[str] = "hvac_mode"
ATTR_TARGET_TEMP_HIGH: Final[str] = "target_temp_high"
ATTR_TARGET_TEMP_LOW: Final[str] = "target_temp_low"

SUPPORT_TARGET_TEMPERATURE: Final[int] = 1
SUPPORT_TARGET_TEMPERATURE_RANGE: Final[int] = 2
SUPPORT_TARGET_HUMIDITY: Final[int] = 4
SUPPORT_FAN_MODE: Final[int] = 8
SUPPORT_PRESET_MODE: Final[int] = 16
SUPPORT_SWING_MODE: Final[int] = 32
SUPPORT_AUX_HEAT: Final[int] = 64

PRECISION_TENTHS: Final[float] = 0.1
# Temperature attribute
ATTR_TEMPERATURE: Final[str] = "temperature"
TEMP_CELSIUS: Final[str] = "°C"


class HVACMode(StrEnum):
    OFF = "off"
    # Heating
    HEAT = "heat"
    # Cooling
    COOL = "cool"
    # The device supports heating/cooling to a range
    HEAT_COOL = "heat_cool"
    # The temperature is set based on a schedule, learned behavior, AI or some
    # other related mechanism. User is not able to adjust the temperature
    AUTO = "auto"
    # Device is in Dry/Humidity mode
    DRY = "dry"
    # Only the fan is on, not fan and another mode like cool
    FAN_ONLY = "fan_only"


class Preset(StrEnum):
    # No preset is active
    NONE = "none"
    # Device is running an energy-saving mode
    ECO = "eco"
    # Device is in away mode
    AWAY = "away"
    # Device turn all valve full up
    BOOST = "boost"
    # Device is in comfort mode
    COMFORT = "comfort"
    # Device is in home mode
    HOME = "home"
    # Device is prepared for sleep
    SLEEP = "sleep"
    # Device is reacting to activity (e.g. movement sensors)
    ACTIVITY = "activity"
    SCHEDULE = "Schedule"
    COMPLEX = "Complex"
    TEMP_MANUAL = "Temporary manual"


class FanState(StrEnum):
    # Possible fan state
    ON = "on"
    OFF = "off"
    AUTO = "auto"
    LOW = "low"
    MEDIUM = "medium"
    HIGH = "high"
    TOP = "top"
    MIDDLE = "middle"
    FOCUS = "focus"
    DIFFUSE = "diffuse"


class CurrentHVAC(StrEnum):
    OFF = "off"
    HEAT = "heating"
    COOL = "cooling"
    DRY = "drying"
    IDLE = "idle"
    FAN = "fan"


RUNNING_MODE = {0x00: HVACMode.OFF, 0x03: HVACMode.COOL, 0x04: HVACMode.HEAT}

SEQ_OF_OPERATION = {
    0x00: (HVACMode.OFF, HVACMode.COOL),  # cooling only
    0x01: (HVACMode.OFF, HVACMode.COOL),  # cooling with reheat
    0x02: (HVACMode.OFF, HVACMode.HEAT),  # heating only
    0x03: (HVACMode.OFF, HVACMode.HEAT),  # heating with reheat
    # cooling and heating 4-pipes
    0x04: (HVACMode.OFF, HVACMode.HEAT_COOL, HVACMode.COOL, HVACMode.HEAT),
    # cooling and heating 4-pipes
    0x05: (HVACMode.OFF, HVACMode.HEAT_COOL, HVACMode.COOL, HVACMode.HEAT),
    0x06: (HVACMode.COOL, HVACMode.HEAT, HVACMode.OFF),  # centralite specific
    0x07: (HVACMode.HEAT_COOL, HVACMode.OFF),  # centralite specific
}

HVAC_MODE_2_SYSTEM = {
    HVACMode.OFF: ThermostatCluster.SystemMode.Off,
    HVACMode.HEAT_COOL: ThermostatCluster.SystemMode.Auto,
    HVACMode.COOL: ThermostatCluster.SystemMode.Cool,
    HVACMode.HEAT: ThermostatCluster.SystemMode.Heat,
    HVACMode.FAN_ONLY: ThermostatCluster.SystemMode.Fan_only,
    HVACMode.DRY: ThermostatCluster.SystemMode.Dry,
}

SYSTEM_MODE_2_HVAC = {
    ThermostatCluster.SystemMode.Off: HVACMode.OFF,
    ThermostatCluster.SystemMode.Auto: HVACMode.HEAT_COOL,
    ThermostatCluster.SystemMode.Cool: HVACMode.COOL,
    ThermostatCluster.SystemMode.Heat: HVACMode.HEAT,
    ThermostatCluster.SystemMode.Emergency_Heating: HVACMode.HEAT,
    ThermostatCluster.SystemMode.Pre_cooling: HVACMode.COOL,  # this is 'precooling'. is it the same?
    ThermostatCluster.SystemMode.Fan_only: HVACMode.FAN_ONLY,
    ThermostatCluster.SystemMode.Dry: HVACMode.DRY,
    ThermostatCluster.SystemMode.Sleep: HVACMode.OFF,
}

ZCL_TEMP: Final[int] = 100
_LOGGER = logging.getLogger(__name__)


"""TODO implement to_json and get_state methods"""


@MULTI_MATCH(
    cluster_handler_names=CLUSTER_HANDLER_THERMOSTAT,
    aux_cluster_handlers=CLUSTER_HANDLER_FAN,
    stop_on_match_group=CLUSTER_HANDLER_THERMOSTAT,
)
class Thermostat(PlatformEntity):

    PLATFORM = Platform.CLIMATE
    DEFAULT_MAX_TEMP = 35
    DEFAULT_MIN_TEMP = 7

    def __init__(
        self,
        unique_id: str,
        cluster_handlers: list[ClusterHandler],
        endpoint: Endpoint,
        device: Device,
    ):
        """Initialize the thermostat."""
        super().__init__(unique_id, cluster_handlers, endpoint, device)
        if CLUSTER_HANDLER_THERMOSTAT not in self.cluster_handlers:
            raise ValueError("No Thermostat cluster handler found")
        self._thermostat_cluster_handler: ClusterHandler = self.cluster_handlers[
            CLUSTER_HANDLER_THERMOSTAT
        ]
        self._preset: Preset = Preset.NONE
        self._presets: list[Preset] = []
        self._supported_flags = SUPPORT_TARGET_TEMPERATURE
        self._fan_cluster_handler: Optional[ClusterHandler] = self.cluster_handlers.get(
            CLUSTER_HANDLER_FAN
        )
        self._thermostat_cluster_handler.on_event(
            CLUSTER_HANDLER_EVENT, self._handle_event_protocol
        )

    @property
    def current_temperature(self) -> Union[float, None]:
        """Return the current temperature."""
        if self._thermostat_cluster_handler.local_temp is None:
            return None
        return self._thermostat_cluster_handler.local_temp / ZCL_TEMP

    @property
    def extra_state_attributes(self) -> dict:
        """Return device specific state attributes."""
        data = {}
        if self.hvac_mode:
            mode = SYSTEM_MODE_2_HVAC.get(
                self._thermostat_cluster_handler.system_mode, "unknown"
            )
            data[
                ATTR_SYS_MODE
            ] = f"[{self._thermostat_cluster_handler.system_mode}]/{mode}"
        if self._thermostat_cluster_handler.occupancy is not None:
            data[ATTR_OCCUPANCY] = self._thermostat_cluster_handler.occupancy
        if self._thermostat_cluster_handler.occupied_cooling_setpoint is not None:
            data[
                ATTR_OCCP_COOL_SETPT
            ] = self._thermostat_cluster_handler.occupied_cooling_setpoint
        if self._thermostat_cluster_handler.occupied_heating_setpoint is not None:
            data[
                ATTR_OCCP_HEAT_SETPT
            ] = self._thermostat_cluster_handler.occupied_heating_setpoint
        if self._thermostat_cluster_handler.pi_heating_demand is not None:
            data[
                ATTR_PI_HEATING_DEMAND
            ] = self._thermostat_cluster_handler.pi_heating_demand
        if self._thermostat_cluster_handler.pi_cooling_demand is not None:
            data[
                ATTR_PI_COOLING_DEMAND
            ] = self._thermostat_cluster_handler.pi_cooling_demand

        unoccupied_cooling_setpoint = (
            self._thermostat_cluster_handler.unoccupied_cooling_setpoint
        )
        if unoccupied_cooling_setpoint is not None:
            data[ATTR_UNOCCP_COOL_SETPT] = unoccupied_cooling_setpoint

        unoccupied_heating_setpoint = (
            self._thermostat_cluster_handler.unoccupied_heating_setpoint
        )
        if unoccupied_heating_setpoint is not None:
            data[ATTR_UNOCCP_HEAT_SETPT] = unoccupied_heating_setpoint
        return data

    @property
    def fan_mode(self) -> Union[str, None]:
        """Return current FAN mode."""
        if self._thermostat_cluster_handler.running_state is None:
            return FanState.AUTO

        if self._thermostat_cluster_handler.running_state & (
            ThermostatCluster.RunningState.Fan_State_On
            | ThermostatCluster.RunningState.Fan_2nd_Stage_On
            | ThermostatCluster.RunningState.Fan_3rd_Stage_On
        ):
            return FanState.ON
        return FanState.AUTO

    @property
    def fan_modes(self) -> Union[list[str], None]:
        """Return supported FAN modes."""
        if not self._fan_cluster_handler:
            return None
        return [FanState.AUTO, FanState.ON]

    @property
    def hvac_action(self) -> Union[str, None]:
        """Return the current HVAC action."""
        if (
            self._thermostat_cluster_handler.pi_heating_demand is None
            and self._thermostat_cluster_handler.pi_cooling_demand is None
        ):
            return self._rm_rs_action
        return self._pi_demand_action

    @property
    def _rm_rs_action(self) -> Union[str, None]:
        """Return the current HVAC action based on running mode and running state."""

        if (running_state := self._thermostat_cluster_handler.running_state) is None:
            return None
        if running_state & (
            ThermostatCluster.RunningState.Heat_State_On
            | ThermostatCluster.RunningState.Heat_2nd_Stage_On
        ):
            return CurrentHVAC.HEAT
        if running_state & (
            ThermostatCluster.RunningState.Cool_State_On
            | ThermostatCluster.RunningState.Cool_2nd_Stage_On
        ):
            return CurrentHVAC.COOL
        if running_state & (
            ThermostatCluster.RunningState.Fan_State_On
            | ThermostatCluster.RunningState.Fan_2nd_Stage_On
            | ThermostatCluster.RunningState.Fan_3rd_Stage_On
        ):
            return CurrentHVAC.FAN
        if running_state & ThermostatCluster.RunningState.Idle:
            return CurrentHVAC.IDLE
        if self.hvac_mode != HVACMode.OFF:
            return CurrentHVAC.IDLE
        return CurrentHVAC.OFF

    @property
    def _pi_demand_action(self) -> Union[str, None]:
        """Return the current HVAC action based on pi_demands."""

        heating_demand = self._thermostat_cluster_handler.pi_heating_demand
        if heating_demand is not None and heating_demand > 0:
            return CurrentHVAC.HEAT
        cooling_demand = self._thermostat_cluster_handler.pi_cooling_demand
        if cooling_demand is not None and cooling_demand > 0:
            return CurrentHVAC.COOL

        if self.hvac_mode != HVACMode.OFF:
            return CurrentHVAC.IDLE
        return CurrentHVAC.OFF

    @property
    def hvac_mode(self) -> Union[str, None]:
        """Return HVAC operation mode."""
        return SYSTEM_MODE_2_HVAC.get(self._thermostat_cluster_handler.system_mode)

    @property
    def hvac_modes(self) -> tuple[str, ...]:
        """Return the list of available HVAC operation modes."""
        return SEQ_OF_OPERATION.get(
            self._thermostat_cluster_handler.ctrl_seqe_of_oper, (HVACMode.OFF,)
        )

    @property
    def precision(self) -> float:
        """Return the precision of the system."""
        return PRECISION_TENTHS

    @property
    def preset_mode(self) -> Union[str, None]:
        """Return current preset mode."""
        return self._preset

    @property
    def preset_modes(self) -> Union[list[str], None]:
        """Return supported preset modes."""
        return self._presets

    @property
    def supported_features(self) -> int:
        """Return the list of supported features."""
        features = self._supported_flags
        if HVACMode.HEAT_COOL in self.hvac_modes:
            features |= SUPPORT_TARGET_TEMPERATURE_RANGE
        if self._fan_cluster_handler is not None:
            self._supported_flags |= SUPPORT_FAN_MODE
        return features

    @property
    def target_temperature(self) -> Union[float, None]:
        """Return the temperature we try to reach."""
        temp = None
        if self.hvac_mode == HVACMode.COOL:
            if self.preset_mode == Preset.AWAY:
                temp = self._thermostat_cluster_handler.unoccupied_cooling_setpoint
            else:
                temp = self._thermostat_cluster_handler.occupied_cooling_setpoint
        elif self.hvac_mode == HVACMode.HEAT:
            if self.preset_mode == Preset.AWAY:
                temp = self._thermostat_cluster_handler.unoccupied_heating_setpoint
            else:
                temp = self._thermostat_cluster_handler.occupied_heating_setpoint
        if temp is None:
            return temp
        return round(temp / ZCL_TEMP, 1)

    @property
    def target_temperature_high(self) -> Union[float, None]:
        """Return the upper bound temperature we try to reach."""
        if self.hvac_mode != HVACMode.HEAT_COOL:
            return None
        if self.preset_mode == Preset.AWAY:
            temp = self._thermostat_cluster_handler.unoccupied_cooling_setpoint
        else:
            temp = self._thermostat_cluster_handler.occupied_cooling_setpoint

        if temp is None:
            return temp

        return round(temp / ZCL_TEMP, 1)

    @property
    def target_temperature_low(self) -> Union[float, None]:
        """Return the lower bound temperature we try to reach."""
        if self.hvac_mode != HVACMode.HEAT_COOL:
            return None
        if self.preset_mode == Preset.AWAY:
            temp = self._thermostat_cluster_handler.unoccupied_heating_setpoint
        else:
            temp = self._thermostat_cluster_handler.occupied_heating_setpoint

        if temp is None:
            return temp
        return round(temp / ZCL_TEMP, 1)

    @property
    def temperature_unit(self) -> str:
        """Return the unit of measurement used by the platform."""
        return TEMP_CELSIUS

    @property
    def max_temp(self) -> float:
        """Return the maximum temperature."""
        temps = []
        if HVACMode.HEAT in self.hvac_modes:
            temps.append(self._thermostat_cluster_handler.max_heat_setpoint_limit)
        if HVACMode.COOL in self.hvac_modes:
            temps.append(self._thermostat_cluster_handler.max_cool_setpoint_limit)

        if not temps:
            return self.DEFAULT_MAX_TEMP
        return round(max(temps) / ZCL_TEMP, 1)

    @property
    def min_temp(self) -> float:
        """Return the minimum temperature."""
        temps = []
        if HVACMode.HEAT in self.hvac_modes:
            temps.append(self._thermostat_cluster_handler.min_heat_setpoint_limit)
        if HVACMode.COOL in self.hvac_modes:
            temps.append(self._thermostat_cluster_handler.min_cool_setpoint_limit)

        if not temps:
            return self.DEFAULT_MIN_TEMP
        return round(min(temps) / ZCL_TEMP, 1)

    async def handle_cluster_handler_attribute_updated(
        self, event: ClusterAttributeUpdatedEvent
    ) -> None:
        """Handle attribute update from device."""
        if (
            event.name in (ATTR_OCCP_COOL_SETPT, ATTR_OCCP_HEAT_SETPT)
            and self.preset_mode == Preset.AWAY
        ):
            # occupancy attribute is an unreportable attribute, but if we get
            # an attribute update for an "occupied" setpoint, there's a chance
            # occupancy has changed
            if await self._thermostat_cluster_handler.get_occupancy() is True:
                self._preset = Preset.NONE

        self.debug("Attribute '%s' = %s update", event.name, event.value)
        self.maybe_send_state_changed_event()

    async def async_set_fan_mode(self, fan_mode: str) -> None:
        """Set fan mode."""
        if fan_mode not in self.fan_modes:
            self.warning("Unsupported '%s' fan mode", fan_mode)
            return

        if fan_mode == FanState.ON:
            mode = FanCluster.FanMode.On
        else:
            mode = FanCluster.FanMode.Auto

        await self._fan_cluster_handler.async_set_speed(mode)

    async def async_set_hvac_mode(self, hvac_mode: str) -> None:
        """Set new target operation mode."""
        if hvac_mode not in self.hvac_modes:
            self.warning(
                "can't set '%s' mode. Supported modes are: %s",
                hvac_mode,
                self.hvac_modes,
            )
            return

        if await self._thermostat_cluster_handler.async_set_operation_mode(
            HVAC_MODE_2_SYSTEM[hvac_mode]
        ):
            self.maybe_send_state_changed_event()

    async def async_set_preset_mode(self, preset_mode: str) -> None:
        """Set new preset mode."""
        if preset_mode not in self.preset_modes:
            self.debug("preset mode '%s' is not supported", preset_mode)
            return

        if (
            self.preset_mode
            not in (
                preset_mode,
                Preset.NONE,
            )
            and not await self.async_preset_handler(self.preset_mode, enable=False)
        ):
            self.debug("Couldn't turn off '%s' preset", self.preset_mode)
            return

        if preset_mode != Preset.NONE and not await self.async_preset_handler(
            preset_mode, enable=True
        ):
            self.debug("Couldn't turn on '%s' preset", preset_mode)
            return
        self._preset = preset_mode
        self.maybe_send_state_changed_event()

    async def async_set_temperature(self, **kwargs: Any) -> None:
        """Set new target temperature."""
        low_temp = kwargs.get(ATTR_TARGET_TEMP_LOW)
        high_temp = kwargs.get(ATTR_TARGET_TEMP_HIGH)
        temp = kwargs.get(ATTR_TEMPERATURE)
        hvac_mode = kwargs.get(ATTR_HVAC_MODE)

        if hvac_mode is not None:
            await self.async_set_hvac_mode(hvac_mode)

        thrm = self._thermostat_cluster_handler
        if self.hvac_mode == HVACMode.HEAT_COOL:
            success = True
            if low_temp is not None:
                low_temp = int(low_temp * ZCL_TEMP)
                success = success and await thrm.async_set_heating_setpoint(
                    low_temp, self.preset_mode == Preset.AWAY
                )
                self.debug("Setting heating %s setpoint: %s", low_temp, success)
            if high_temp is not None:
                high_temp = int(high_temp * ZCL_TEMP)
                success = success and await thrm.async_set_cooling_setpoint(
                    high_temp, self.preset_mode == Preset.AWAY
                )
                self.debug("Setting cooling %s setpoint: %s", low_temp, success)
        elif temp is not None:
            temp = int(temp * ZCL_TEMP)
            if self.hvac_mode == HVACMode.COOL:
                success = await thrm.async_set_cooling_setpoint(
                    temp, self.preset_mode == Preset.AWAY
                )
            elif self.hvac_mode == HVACMode.HEAT:
                success = await thrm.async_set_heating_setpoint(
                    temp, self.preset_mode == Preset.AWAY
                )
            else:
                self.debug("Not setting temperature for '%s' mode", self.hvac_mode)
                return
        else:
            self.debug("incorrect %s setting for '%s' mode", kwargs, self.hvac_mode)
            return

        if success:
            self.maybe_send_state_changed_event()

    async def async_preset_handler(self, preset: str, enable: bool = False) -> bool:
        """Set the preset mode via handler."""

        handler = getattr(self, f"async_preset_handler_{preset}")
        return await handler(enable)


@MULTI_MATCH(
    cluster_handler_names={CLUSTER_HANDLER_THERMOSTAT, "sinope_manufacturer_specific"},
    manufacturers="Sinope Technologies",
    stop_on_match_group=CLUSTER_HANDLER_THERMOSTAT,
)
class SinopeTechnologiesThermostat(Thermostat):
    """Sinope Technologies Thermostat."""

    manufacturer = 0x119C
    update_time_interval = (2700, 4500)

    def __init__(
        self,
        unique_id: str,
        cluster_handlers: list[ClusterHandler],
        endpoint: Endpoint,
        device: Device,
    ):
        """Initialize the thermostat."""
        super().__init__(unique_id, cluster_handlers, endpoint, device)
        self._presets: list[Preset] = [Preset.AWAY, Preset.NONE]
        self._supported_flags |= SUPPORT_PRESET_MODE
        self._manufacturer_ch: ClusterHandler = self.cluster_handlers[
            "sinope_manufacturer_specific"
        ]

        @periodic(self.update_time_interval)
        async def _update_time() -> None:
            await self._async_update_time()

        self._update_time_task = asyncio.create_task(_update_time())

    @property
    def _rm_rs_action(self) -> Union[str, None]:
        """Return the current HVAC action based on running mode and running state."""

        running_mode = self._thermostat_cluster_handler.running_mode
        if running_mode == ThermostatCluster.SystemMode.Heat:
            return CurrentHVAC.HEAT
        if running_mode == ThermostatCluster.SystemMode.Cool:
            return CurrentHVAC.COOL

        running_state = self._thermostat_cluster_handler.running_state
        if running_state and running_state & (
            ThermostatCluster.RunningState.Fan_State_On
            | ThermostatCluster.RunningState.Fan_2nd_Stage_On
            | ThermostatCluster.RunningState.Fan_3rd_Stage_On
        ):
            return CurrentHVAC.FAN
        if (
            self.hvac_mode != HVACMode.OFF
            and running_mode == ThermostatCluster.SystemMode.Off
        ):
            return CurrentHVAC.IDLE
        return CurrentHVAC.OFF

    async def _async_update_time(self) -> None:
        """Update thermostat's time display."""

        secs_2k = (
            datetime.now(datetime.timezone.utc).replace(tzinfo=None)
            - datetime(2000, 1, 1, 0, 0, 0, 0)
        ).total_seconds()

        self.debug("Updating time: %s", secs_2k)
        try:
            await self._manufacturer_ch.cluster.write_attributes(
                {"secs_since_2k": secs_2k}, manufacturer=self.manufacturer
            )
        except (
            Exception,
            asyncio.CancelledError,
        ) as exc:  # pylint: disable=broad-except
            # Do not print the wrapper in the traceback
            self.warning("Error updating time: %s", exc, exc_info=exc)

    async def async_preset_handler_away(self, is_away: bool = False) -> bool:
        """Set occupancy."""
        mfg_code = self._device.manufacturer_code
        res = await self._thermostat_cluster_handler.write_attributes(
            {"set_occupancy": 0 if is_away else 1}, manufacturer=mfg_code
        )

        self.debug("set occupancy to %s. Status: %s", 0 if is_away else 1, res)
        return res


@MULTI_MATCH(
    cluster_handler_names=CLUSTER_HANDLER_THERMOSTAT,
    aux_cluster_handlers=CLUSTER_HANDLER_FAN,
    manufacturers="Zen Within",
    stop_on_match_group=CLUSTER_HANDLER_THERMOSTAT,
)
class ZenWithinThermostat(Thermostat):
    """Zen Within Thermostat implementation."""


@MULTI_MATCH(
    cluster_handler_names=CLUSTER_HANDLER_THERMOSTAT,
    aux_cluster_handlers=CLUSTER_HANDLER_FAN,
    manufacturers="Centralite",
    models={"3157100", "3157100-E"},
    stop_on_match_group=CLUSTER_HANDLER_THERMOSTAT,
)
class CentralitePearl(ZenWithinThermostat):
    """Centralite Pearl Thermostat implementation."""


@STRICT_MATCH(
    cluster_handler_names=CLUSTER_HANDLER_THERMOSTAT,
    manufacturers={
        "_TZE200_ckud7u2l",
        "_TZE200_ywdxldoj",
        "_TZE200_cwnjrr72",
        "_TZE200_b6wax7g0",
        "_TZE200_2atgpdho",
        "_TZE200_pvvbommb",
        "_TZE200_4eeyebrt",
        "_TYST11_ckud7u2l",
        "_TYST11_ywdxldoj",
        "_TYST11_cwnjrr72",
        "_TYST11_2atgpdho",
    },
)
class MoesThermostat(Thermostat):
    """Moes Thermostat implementation."""

    def __init__(
        self,
        unique_id: str,
        cluster_handlers: list[ClusterHandler],
        endpoint: Endpoint,
        device: Device,
    ):
        """Initialize the thermostat."""
        super().__init__(unique_id, cluster_handlers, endpoint, device)
        self._presets: list[Preset] = [
            Preset.NONE,
            Preset.AWAY,
            Preset.SCHEDULE,
            Preset.COMFORT,
            Preset.ECO,
            Preset.BOOST,
            Preset.COMPLEX,
        ]
        self._supported_flags |= SUPPORT_PRESET_MODE

    @property
    def hvac_modes(self) -> tuple[str, ...]:
        """Return only the heat mode, because the device can't be turned off."""
        return (HVACMode.HEAT,)

    async def handle_cluster_handler_attribute_updated(
        self, event: ClusterAttributeUpdatedEvent
    ) -> None:
        """Handle attribute update from device."""
        if event.name == "operation_preset":
            if event.value == 0:
                self._preset = Preset.AWAY
            if event.value == 1:
                self._preset = Preset.SCHEDULE
            if event.value == 2:
                self._preset = Preset.NONE
            if event.value == 3:
                self._preset = Preset.COMFORT
            if event.value == 4:
                self._preset = Preset.ECO
            if event.value == 5:
                self._preset = Preset.BOOST
            if event.value == 6:
                self._preset = Preset.COMPLEX
        await super().handle_cluster_handler_attribute_updated(event)

    async def async_preset_handler(self, preset: str, enable: bool = False) -> bool:
        """Set the preset mode."""
        mfg_code = self._device.manufacturer_code
        if not enable:
            return await self._thermostat_cluster_handler.write_attributes(
                {"operation_preset": 2}, manufacturer=mfg_code
            )
        if preset == Preset.AWAY:
            return await self._thermostat_cluster_handler.write_attributes(
                {"operation_preset": 0}, manufacturer=mfg_code
            )
        if preset == Preset.SCHEDULE:
            return await self._thermostat_cluster_handler.write_attributes(
                {"operation_preset": 1}, manufacturer=mfg_code
            )
        if preset == Preset.COMFORT:
            return await self._thermostat_cluster_handler.write_attributes(
                {"operation_preset": 3}, manufacturer=mfg_code
            )
        if preset == Preset.ECO:
            return await self._thermostat_cluster_handler.write_attributes(
                {"operation_preset": 4}, manufacturer=mfg_code
            )
        if preset == Preset.BOOST:
            return await self._thermostat_cluster_handler.write_attributes(
                {"operation_preset": 5}, manufacturer=mfg_code
            )
        if preset == Preset.COMPLEX:
            return await self._thermostat_cluster_handler.write_attributes(
                {"operation_preset": 6}, manufacturer=mfg_code
            )

        return False


@STRICT_MATCH(
    cluster_handler_names=CLUSTER_HANDLER_THERMOSTAT,
    manufacturers={
        "_TZE200_b6wax7g0",
    },
)
class BecaThermostat(Thermostat):
    """Beca Thermostat implementation."""

    def __init__(
        self,
        unique_id: str,
        cluster_handlers: list[ClusterHandler],
        endpoint: Endpoint,
        device: Device,
    ):
        """Initialize the thermostat."""
        super().__init__(unique_id, cluster_handlers, endpoint, device)
        self._presets: list[Preset] = [
            Preset.NONE,
            Preset.AWAY,
            Preset.SCHEDULE,
            Preset.ECO,
            Preset.BOOST,
            Preset.TEMP_MANUAL,
        ]
        self._supported_flags |= SUPPORT_PRESET_MODE

    @property
    def hvac_modes(self) -> tuple[str, ...]:
        """Return only the heat mode, because the device can't be turned off."""
        return (HVACMode.HEAT,)

    async def handle_cluster_handler_attribute_updated(
        self, event: ClusterAttributeUpdatedEvent
    ) -> None:
        """Handle attribute update from device."""
        if event.name == "operation_preset":
            if event.value == 0:
                self._preset = Preset.AWAY
            if event.value == 1:
                self._preset = Preset.SCHEDULE
            if event.value == 2:
                self._preset = Preset.NONE
            if event.value == 4:
                self._preset = Preset.ECO
            if event.value == 5:
                self._preset = Preset.BOOST
            if event.value == 7:
                self._preset = Preset.TEMP_MANUAL
        await super().handle_cluster_handler_attribute_updated(event)

    async def async_preset_handler(self, preset: str, enable: bool = False) -> bool:
        """Set the preset mode."""
        mfg_code = self._device.manufacturer_code
        if not enable:
            return await self._thermostat_cluster_handler.write_attributes(
                {"operation_preset": 2}, manufacturer=mfg_code
            )
        if preset == Preset.AWAY:
            return await self._thermostat_cluster_handler.write_attributes(
                {"operation_preset": 0}, manufacturer=mfg_code
            )
        if preset == Preset.SCHEDULE:
            return await self._thermostat_cluster_handler.write_attributes(
                {"operation_preset": 1}, manufacturer=mfg_code
            )
        if preset == Preset.ECO:
            return await self._thermostat_cluster_handler.write_attributes(
                {"operation_preset": 4}, manufacturer=mfg_code
            )
        if preset == Preset.BOOST:
            return await self._thermostat_cluster_handler.write_attributes(
                {"operation_preset": 5}, manufacturer=mfg_code
            )
        if preset == Preset.TEMP_MANUAL:
            return await self._thermostat_cluster_handler.write_attributes(
                {"operation_preset": 7}, manufacturer=mfg_code
            )

        return False



"""Websocket API for zhawss."""
from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING, Any, Mapping, TypeVar, cast

import voluptuous as vol
from zigpy.config import CONF_DEVICE, CONF_DEVICE_PATH
from zigpy.types.named import EUI64

from zhaws.server.const import (
    COMMAND,
    CONF_BAUDRATE,
    CONF_DATABASE,
    CONF_ENABLE_QUIRKS,
    CONF_FLOWCONTROL,
    CONF_RADIO_TYPE,
    DEVICES,
    DURATION,
    GROUPS,
    IEEE,
    APICommands,
)
from zhaws.server.websocket.api import decorators, register_api_command
from zhaws.server.zigbee.controller import Controller
from zhaws.server.zigbee.device import (
    ATTR_ATTRIBUTE,
    ATTR_CLUSTER_ID,
    ATTR_CLUSTER_TYPE,
    ATTR_ENDPOINT_ID,
    ATTR_MANUFACTURER_CODE,
    ATTR_VALUE,
    Device,
)
from zhaws.server.zigbee.group import Group, GroupMemberReference

if TYPE_CHECKING:
    from zhaws.server.websocket.client import Client
    from zhaws.server.websocket.server import Server

GROUP = "group"
GROUP_ID = "group_id"
GROUP_IDS = "group_ids"
GROUP_NAME = "group_name"
ATTR_MEMBERS = "members"
MFG_CLUSTER_ID_START = 0xFC00

_LOGGER = logging.getLogger(__name__)

positive_int = vol.All(vol.Coerce(int), vol.Range(min=0))

T = TypeVar("T")


def ensure_list(value: T | None) -> list[T] | list[Any]:
    """Wrap value in list if it is not one."""
    if value is None:
        return []
    return cast("list[T]", value) if isinstance(value, list) else [value]


@decorators.websocket_command(
    {
        vol.Required(COMMAND): str(APICommands.START_NETWORK),
        vol.Required(CONF_RADIO_TYPE): str,
        vol.Required(CONF_DEVICE): vol.Schema(
            {
                vol.Required(CONF_DEVICE_PATH): str,
                vol.Optional(CONF_FLOWCONTROL): str,
                vol.Required(CONF_BAUDRATE): decorators.POSITIVE_INT,
            }
        ),
        vol.Required(CONF_DATABASE): str,
        vol.Required(CONF_ENABLE_QUIRKS): bool,
    }
)
@decorators.async_response
async def start_network(
    server: Server, client: Client, message: dict[str, Any]
) -> None:
    """Start the Zigbee network."""
    await server.controller.start_network(message)
    client.send_result_success(message)


@decorators.websocket_command(
    {
        vol.Required(COMMAND): str(APICommands.STOP_NETWORK),
    }
)
@decorators.async_response
async def stop_network(server: Server, client: Client, message: dict[str, Any]) -> None:
    """Stop the Zigbee network."""
    await server.controller.stop_network()
    client.send_result_success(message)


@decorators.websocket_command(
    {
        vol.Required(COMMAND): str(APICommands.UPDATE_NETWORK_TOPOLOGY),
    }
)
@decorators.async_response
async def update_topology(
    server: Server, client: Client, message: dict[str, Any]
) -> None:
    """Update the Zigbee network topology."""
    await server.controller.application_controller.topology.scan()
    client.send_result_success(message)


@decorators.websocket_command(
    {
        vol.Required(COMMAND): str(APICommands.GET_DEVICES),
    }
)
@decorators.async_response
async def get_devices(server: Server, client: Client, message: dict[str, Any]) -> None:
    """Get Zigbee devices."""
    response_devices: dict[str, dict] = {
        str(ieee): device.zha_device_info
        for ieee, device in server.controller.devices.items()
    }
    _LOGGER.info("devices: %s", response_devices)
    client.send_result_success(message, {DEVICES: response_devices})


@decorators.websocket_command(
    {
        vol.Required(COMMAND): str(APICommands.RECONFIGURE_DEVICE),
        vol.Required(IEEE): EUI64.convert,
    }
)
@decorators.async_response
async def reconfigure_device(
    server: Server, client: Client, message: dict[str, Any]
) -> None:
    """Reconfigure a zigbee device."""
    device = server.controller.devices.get(message[IEEE])
    if device:
        await device.async_configure()
    client.send_result_success(message)


@decorators.websocket_command(
    {
        vol.Required(COMMAND): str(APICommands.GET_GROUPS),
    }
)
@decorators.async_response
async def get_groups(server: Server, client: Client, message: dict[str, Any]) -> None:
    """Get Zigbee groups."""
    groups: dict[int, Any] = {
        id: group.to_json() for id, group in server.controller.groups.items()
    }
    _LOGGER.info("groups: %s", groups)
    client.send_result_success(message, {GROUPS: groups})


@decorators.websocket_command(
    {
        vol.Required(COMMAND): str(APICommands.PERMIT_JOINING),
        vol.Optional(DURATION, default=60): vol.All(vol.Coerce(int), vol.Range(0, 254)),
        vol.Optional(IEEE, default=None): EUI64.convert,
    }
)
@decorators.async_response
async def permit_joining(
    server: Server, client: Client, message: dict[str, Any]
) -> None:
    """Permit joining devices to the Zigbee network."""
    # TODO add permit with code support
    await server.controller.application_controller.permit(
        message[DURATION], message[IEEE]
    )
    client.send_result_success(
        message,
        {DURATION: message[DURATION]},
    )


@decorators.websocket_command(
    {
        vol.Required(COMMAND): str(APICommands.REMOVE_DEVICE),
        vol.Required(IEEE): str,
    }
)
@decorators.async_response
async def remove_device(
    server: Server, client: Client, message: dict[str, Any]
) -> None:
    """Permit joining devices to the Zigbee network."""
    await server.controller.application_controller.remove(
        EUI64.convert(message["ieee"])
    )
    client.send_result_success(message)


@decorators.websocket_command(
    {
        vol.Required(COMMAND): str(APICommands.READ_CLUSTER_ATTRIBUTES),
        vol.Required(IEEE): EUI64.convert,
        vol.Required(ATTR_ENDPOINT_ID): int,
        vol.Required(ATTR_CLUSTER_ID): int,
        vol.Required(ATTR_CLUSTER_TYPE): str,
        vol.Required("attributes"): vol.All(ensure_list, [str]),
        vol.Optional(ATTR_MANUFACTURER_CODE): int,
    }
)
@decorators.async_response
async def read_cluster_attributes(
    server: Server, client: Client, message: dict[str, Any]
) -> None:
    """Read the specified cluster attributes."""
    device: Device = server.controller.devices[message[IEEE]]
    if not device:
        client.send_result_error(
            message,
            "Device not found",
            f"Device with ieee: {message[IEEE]} not found",
        )
        return
    endpoint_id = message[ATTR_ENDPOINT_ID]
    cluster_id = message[ATTR_CLUSTER_ID]
    cluster_type = message[ATTR_CLUSTER_TYPE]
    attributes = message["attributes"]
    manufacturer = message.get(ATTR_MANUFACTURER_CODE)
    if cluster_id >= MFG_CLUSTER_ID_START and manufacturer is None:
        manufacturer = device.manufacturer_code
    cluster = device.async_get_cluster(
        endpoint_id, cluster_id, cluster_type=cluster_type
    )
    if not cluster:
        client.send_result_error(
            message,
            "Cluster not found",
            f"Cluster: {endpoint_id}:{message[ATTR_CLUSTER_ID]} not found on device with ieee: {message[IEEE]} not found",
        )
        return
    success, failure = await cluster.read_attributes(
        attributes, allow_cache=False, only_cache=False, manufacturer=manufacturer
    )
    client.send_result_success(
        message,
        {
            "device": {
                "ieee": str(message[IEEE]),
            },
            "cluster": {
                "id": cluster.cluster_id,
                "endpoint_id": cluster.endpoint.endpoint_id,
                "name": cluster.name,
                "endpoint_attribute": cluster.ep_attribute,
            },
            "manufacturer_code": manufacturer,
            "succeeded": success,
            "failed": failure,
        },
    )


@decorators.websocket_command(
    {
        vol.Required(COMMAND): str(APICommands.WRITE_CLUSTER_ATTRIBUTE),
        vol.Required(IEEE): EUI64.convert,
        vol.Required(ATTR_ENDPOINT_ID): int,
        vol.Required(ATTR_CLUSTER_ID): int,
        vol.Required(ATTR_CLUSTER_TYPE): str,
        vol.Required(ATTR_ATTRIBUTE): str,
        vol.Required(ATTR_VALUE): vol.Any(str, int, float, bool),
        vol.Optional(ATTR_MANUFACTURER_CODE): int,
    }
)
@decorators.async_response
async def write_cluster_attribute(
    server: Server, client: Client, message: dict[str, Any]
) -> None:
    """Set the value of the specifiec cluster attribute."""
    device: Device = server.controller.devices[message[IEEE]]
    if not device:
        client.send_result_error(
            message,
            "Device not found",
            f"Device with ieee: {message[IEEE]} not found",
        )
        return
    endpoint_id = message[ATTR_ENDPOINT_ID]
    cluster_id = message[ATTR_CLUSTER_ID]
    cluster_type = message[ATTR_CLUSTER_TYPE]
    attribute = message[ATTR_ATTRIBUTE]
    value = message[ATTR_VALUE]
    manufacturer = message.get(ATTR_MANUFACTURER_CODE)
    if cluster_id >= MFG_CLUSTER_ID_START and manufacturer is None:
        manufacturer = device.manufacturer_code
    cluster = device.async_get_cluster(
        endpoint_id, cluster_id, cluster_type=cluster_type
    )
    if not cluster:
        client.send_result_error(
            message,
            "Cluster not found",
            f"Cluster: {endpoint_id}:{message[ATTR_CLUSTER_ID]} not found on device with ieee: {message[IEEE]} not found",
        )
        return
    response = await device.write_zigbee_attribute(
        endpoint_id,
        cluster_id,
        attribute,
        value,
        cluster_type=cluster_type,
        manufacturer=manufacturer,
    )
    client.send_result_success(
        message,
        {
            "device": {
                "ieee": str(message[IEEE]),
            },
            "cluster": {
                "id": cluster.cluster_id,
                "endpoint_id": cluster.endpoint.endpoint_id,
                "name": cluster.name,
                "endpoint_attribute": cluster.ep_attribute,
            },
            "manufacturer_code": manufacturer,
            "response": {
                "attribute": attribute,
                "status": response[0][0].status.name,  # type: ignore
            },  # TODO there has to be a better way to do this
        },
    )


def cv_group_member(value: Any) -> GroupMemberReference:
    """Validate and transform a group member."""
    if not isinstance(value, Mapping):
        raise vol.Invalid("Not a group member")
    try:
        group_member = GroupMemberReference(
            ieee=EUI64.convert(value["ieee"]), endpoint_id=value["endpoint_id"]
        )
    except KeyError as err:
        raise vol.Invalid("Not a group member") from err

    return group_member


@decorators.websocket_command(
    {
        vol.Required(COMMAND): str(APICommands.CREATE_GROUP),
        vol.Required(GROUP_NAME): str,
        vol.Optional(GROUP_ID): positive_int,
        vol.Optional(ATTR_MEMBERS): vol.All(ensure_list, [cv_group_member]),
    }
)
@decorators.async_response
async def create_group(server: Server, client: Client, message: dict[str, Any]) -> None:
    """create a new group."""
    controller: Controller = server.controller
    group_name = message[GROUP_NAME]
    members = cast(list[GroupMemberReference], message.get(ATTR_MEMBERS))
    group_id = message.get(GROUP_ID)
    group: Group = await controller.async_create_zigpy_group(
        group_name, members, group_id
    )
    client.send_result_success(message, {"group": group.to_json()})


@decorators.websocket_command(
    {
        vol.Required(COMMAND): str(APICommands.REMOVE_GROUPS),
        vol.Required(GROUP_IDS): vol.All(ensure_list, [positive_int]),
    }
)
@decorators.async_response
async def remove_groups(
    server: Server, client: Client, message: dict[str, Any]
) -> None:
    """Remove the specified groups."""
    controller: Controller = server.controller
    group_ids = message[GROUP_IDS]

    if len(group_ids) > 1:
        tasks = []
        for group_id in group_ids:
            tasks.append(controller.async_remove_zigpy_group(group_id))
        await asyncio.gather(*tasks)
    else:
        await controller.async_remove_zigpy_group(group_ids[0])
    groups: dict[int, Any] = {
        id: group.to_json() for id, group in server.controller.groups.items()
    }
    client.send_result_success(message, {GROUPS: groups})


@decorators.websocket_command(
    {
        vol.Required(COMMAND): str(APICommands.ADD_GROUP_MEMBERS),
        vol.Required(GROUP_ID): positive_int,
        vol.Required(ATTR_MEMBERS): vol.All(ensure_list, [cv_group_member]),
    }
)
@decorators.async_response
async def add_group_members(
    server: Server, client: Client, message: dict[str, Any]
) -> None:
    """Add members to a ZHA group."""
    controller: Controller = server.controller
    group_id = message[GROUP_ID]
    members = cast(list[GroupMemberReference], message.get(ATTR_MEMBERS))
    group = None

    if group_id in controller.groups:
        group = controller.groups[group_id]
        await group.async_add_members(members)
    if not group:
        client.send_result_error(message, "G1", "ZHA Group not found")
        return
    ret_group = group.to_json()
    client.send_result_success(message, {GROUP: ret_group})


@decorators.websocket_command(
    {
        vol.Required(COMMAND): str(APICommands.REMOVE_GROUP_MEMBERS),
        vol.Required(GROUP_ID): positive_int,
        vol.Required(ATTR_MEMBERS): vol.All(ensure_list, [cv_group_member]),
    }
)
@decorators.async_response
async def remove_group_members(
    server: Server, client: Client, message: dict[str, Any]
) -> None:
    """Remove members from a ZHA group."""
    controller: Controller = server.controller
    group_id = message[GROUP_ID]
    members = cast(list[GroupMemberReference], message.get(ATTR_MEMBERS))
    group = None

    if group_id in controller.groups:
        group = controller.groups[group_id]
        await group.async_remove_members(members)
    if not group:
        client.send_result_error(message, "G1", "ZHA Group not found")
        return
    ret_group = group.to_json()
    client.send_result_success(message, {GROUP: ret_group})


def load_api(server: Server) -> None:
    """Load the api command handlers."""
    register_api_command(server, start_network)
    register_api_command(server, stop_network)
    register_api_command(server, get_devices)
    register_api_command(server, get_groups)
    register_api_command(server, create_group)
    register_api_command(server, remove_groups)
    register_api_command(server, add_group_members)
    register_api_command(server, remove_group_members)
    register_api_command(server, permit_joining)
    register_api_command(server, remove_device)
    register_api_command(server, update_topology)
    register_api_command(server, read_cluster_attributes)
    register_api_command(server, write_cluster_attribute)



"""Binary Sensor module for zhawss."""
from __future__ import annotations

import functools
from typing import TYPE_CHECKING, Union

from zhaws.server.platforms import PlatformEntity
from zhaws.server.platforms.registries import PLATFORM_ENTITIES, Platform
from zhaws.server.zigbee.cluster import (
    CLUSTER_HANDLER_EVENT,
    ClusterAttributeUpdatedEvent,
)
from zhaws.server.zigbee.cluster.const import (
    CLUSTER_HANDLER_ACCELEROMETER,
    CLUSTER_HANDLER_BINARY_INPUT,
    CLUSTER_HANDLER_OCCUPANCY,
    CLUSTER_HANDLER_ON_OFF,
    CLUSTER_HANDLER_ZONE,
)

if TYPE_CHECKING:
    from zhaws.server.zigbee.cluster import ClusterHandler
    from zhaws.server.zigbee.device import Device
    from zhaws.server.zigbee.endpoint import Endpoint

STRICT_MATCH = functools.partial(PLATFORM_ENTITIES.strict_match, Platform.BINARY_SENSOR)
MULTI_MATCH = functools.partial(
    PLATFORM_ENTITIES.multipass_match, Platform.BINARY_SENSOR
)


class BinarySensor(PlatformEntity):
    """BinarySensor platform entity."""

    SENSOR_ATTR: Union[str, None] = None
    PLATFORM: Platform = Platform.BINARY_SENSOR

    def __init__(
        self,
        unique_id: str,
        cluster_handlers: list[ClusterHandler],
        endpoint: Endpoint,
        device: Device,
    ):
        """Initialize the binary sensor."""
        super().__init__(unique_id, cluster_handlers, endpoint, device)
        self._cluster_handler: ClusterHandler = cluster_handlers[0]
        self._cluster_handler.on_event(
            CLUSTER_HANDLER_EVENT, self._handle_event_protocol
        )
        self._state: bool = bool(self._cluster_handler.cluster.get(self.SENSOR_ATTR))

    @property
    def is_on(self) -> bool:
        """Return True if the binary sensor is on."""
        return self._state

    def get_state(self) -> dict:
        """Return the state of the binary sensor."""
        response = super().get_state()
        response["state"] = self.is_on
        return response

    def handle_cluster_handler_attribute_updated(
        self, event: ClusterAttributeUpdatedEvent
    ) -> None:
        """handle attribute updates from the cluster handler."""
        if self.SENSOR_ATTR is None or self.SENSOR_ATTR != event.name:
            return
        self._state = bool(event.value)
        self.maybe_send_state_changed_event()

    async def async_update(self) -> None:
        """Attempt to retrieve on off state from the binary sensor."""
        await super().async_update()
        attribute = getattr(self._cluster_handler, "value_attribute", "on_off")
        attr_value = await self._cluster_handler.get_attribute_value(attribute)
        if attr_value is not None:
            self._state = attr_value
            self.maybe_send_state_changed_event()

    def to_json(self) -> dict:
        """Return a JSON representation of the binary sensor."""
        json = super().to_json()
        json["sensor_attribute"] = self.SENSOR_ATTR
        return json


@MULTI_MATCH(cluster_handler_names=CLUSTER_HANDLER_ACCELEROMETER)
class Accelerometer(BinarySensor):
    """ZHA BinarySensor."""

    SENSOR_ATTR = "acceleration"


@MULTI_MATCH(cluster_handler_names=CLUSTER_HANDLER_OCCUPANCY)
class Occupancy(BinarySensor):
    """ZHA BinarySensor."""

    SENSOR_ATTR = "occupancy"


@STRICT_MATCH(cluster_handler_names=CLUSTER_HANDLER_ON_OFF)
class Opening(BinarySensor):
    """ZHA BinarySensor."""

    SENSOR_ATTR = "on_off"


@MULTI_MATCH(cluster_handler_names=CLUSTER_HANDLER_BINARY_INPUT)
class BinaryInput(BinarySensor):
    """ZHA BinarySensor."""

    SENSOR_ATTR = "present_value"


@STRICT_MATCH(
    cluster_handler_names=CLUSTER_HANDLER_ON_OFF,
    manufacturers="IKEA of Sweden",
    models=lambda model: isinstance(model, str)
    and model is not None
    and model.find("motion") != -1,
)
@STRICT_MATCH(
    cluster_handler_names=CLUSTER_HANDLER_ON_OFF,
    manufacturers="Philips",
    models={"SML001", "SML002"},
)
class Motion(BinarySensor):
    """ZHA BinarySensor."""

    SENSOR_ATTR = "on_off"


@MULTI_MATCH(cluster_handler_names=CLUSTER_HANDLER_ZONE)
class IASZone(BinarySensor):
    """ZHA IAS BinarySensor."""

    SENSOR_ATTR = "zone_status"

    def __init__(
        self,
        unique_id: str,
        cluster_handlers: list[ClusterHandler],
        endpoint: Endpoint,
        device: Device,
    ):
        """Initialize the binary sensor."""
        super().__init__(unique_id, cluster_handlers, endpoint, device)
        value = self._cluster_handler.cluster.get(self.SENSOR_ATTR)
        self._state: bool = value & 3 if value is not None else False

    async def async_update(self) -> None:
        """Attempt to retrieve on off state from the binary sensor."""
        await super().async_update()
        value = await self._cluster_handler.get_attribute_value("zone_status")
        if value is not None:
            self._state = value & 3
            self.maybe_send_state_changed_event()

    def to_json(self) -> dict:
        """Return a JSON representation of the binary sensor."""
        json = super().to_json()
        json["zone_type"] = self._cluster_handler.cluster.get("zone_type")
        return json



"""WS api for the lock platform entity."""
from __future__ import annotations

from typing import TYPE_CHECKING, Any

import voluptuous as vol

from zhaws.backports.enum import StrEnum
from zhaws.server.platforms.api import (
    execute_platform_entity_command,
    platform_entity_command_schema,
)
from zhaws.server.websocket.api import decorators, register_api_command

if TYPE_CHECKING:
    from zhaws.server.websocket.client import Client
    from zhaws.server.websocket.server import Server


class LockCommands(StrEnum):
    """Lock commands."""

    LOCK = "lock_lock"
    UNLOCK = "lock_unlock"
    SET_USER_LOCK_CODE = "lock_set_user_lock_code"
    ENABLE_USER_LOCK_CODE = "lock_enable_user_lock_code"
    DISABLE_USER_LOCK_CODE = "lock_disable_user_lock_code"
    CLEAR_USER_LOCK_CODE = "lock_clear_user_lock_code"


@decorators.websocket_command(platform_entity_command_schema(LockCommands.LOCK))
@decorators.async_response
async def lock(server: Server, client: Client, message: dict[str, Any]) -> None:
    """Lock the lock."""
    await execute_platform_entity_command(server, client, message, "async_lock")


@decorators.websocket_command(platform_entity_command_schema(LockCommands.UNLOCK))
@decorators.async_response
async def unlock(server: Server, client: Client, message: dict[str, Any]) -> None:
    """Unlock the lock."""
    await execute_platform_entity_command(server, client, message, "async_unlock")


@decorators.websocket_command(
    platform_entity_command_schema(
        LockCommands.SET_USER_LOCK_CODE,
        {
            vol.Required("code_slot"): vol.Coerce(int),
            vol.Required("user_code"): str,
        },
    )
)
@decorators.async_response
async def set_user_lock_code(
    server: Server, client: Client, message: dict[str, Any]
) -> None:
    """Set a user lock code in the specified slot for the lock."""
    await execute_platform_entity_command(
        server, client, message, "async_set_user_lock_code"
    )


@decorators.websocket_command(
    platform_entity_command_schema(
        LockCommands.ENABLE_USER_LOCK_CODE,
        {
            vol.Required("code_slot"): vol.Coerce(int),
        },
    )
)
@decorators.async_response
async def enable_user_lock_code(
    server: Server, client: Client, message: dict[str, Any]
) -> None:
    """Enable a user lock code for the lock."""
    await execute_platform_entity_command(
        server, client, message, "async_enable_user_lock_code"
    )


@decorators.websocket_command(
    platform_entity_command_schema(
        LockCommands.DISABLE_USER_LOCK_CODE,
        {
            vol.Required("code_slot"): vol.Coerce(int),
        },
    )
)
@decorators.async_response
async def disable_user_lock_code(
    server: Server, client: Client, message: dict[str, Any]
) -> None:
    """Disable a user lock code for the lock."""
    await execute_platform_entity_command(
        server, client, message, "async_disable_user_lock_code"
    )


@decorators.websocket_command(
    platform_entity_command_schema(
        LockCommands.CLEAR_USER_LOCK_CODE,
        {
            vol.Required("code_slot"): vol.Coerce(int),
        },
    )
)
@decorators.async_response
async def clear_user_lock_code(
    server: Server, client: Client, message: dict[str, Any]
) -> None:
    """Clear a user lock code for the lock."""
    await execute_platform_entity_command(
        server, client, message, "async_clear_user_lock_code"
    )


def load_api(server: Server) -> None:
    """Load the api command handlers."""
    register_api_command(server, lock)
    register_api_command(server, unlock)
    register_api_command(server, set_user_lock_code)
    register_api_command(server, enable_user_lock_code)
    register_api_command(server, disable_user_lock_code)
    register_api_command(server, clear_user_lock_code)



"""Constants."""

from typing import Final

import voluptuous

from zhaws.backports.enum import StrEnum


class APICommands(StrEnum):
    """WS API commands."""

    # Device commands
    GET_DEVICES = "get_devices"
    REMOVE_DEVICE = "remove_device"
    RECONFIGURE_DEVICE = "reconfigure_device"
    READ_CLUSTER_ATTRIBUTES = "read_cluster_attributes"
    WRITE_CLUSTER_ATTRIBUTE = "write_cluster_attribute"

    # Zigbee API commands
    PERMIT_JOINING = "permit_joining"
    START_NETWORK = "start_network"
    STOP_NETWORK = "stop_network"
    UPDATE_NETWORK_TOPOLOGY = "update_network_topology"

    # Group commands
    GET_GROUPS = "get_groups"
    CREATE_GROUP = "create_group"
    REMOVE_GROUPS = "remove_groups"
    ADD_GROUP_MEMBERS = "add_group_members"
    REMOVE_GROUP_MEMBERS = "remove_group_members"

    # Server API commands
    STOP_SERVER = "stop_server"

    # Light API commands
    LIGHT_TURN_ON = "light_turn_on"
    LIGHT_TURN_OFF = "light_turn_off"

    # Switch API commands
    SWITCH_TURN_ON = "switch_turn_on"
    SWITCH_TURN_OFF = "switch_turn_off"

    SIREN_TURN_ON = "siren_turn_on"
    SIREN_TURN_OFF = "siren_turn_off"

    LOCK_UNLOCK = "lock_unlock"
    LOCK_LOCK = "lock_lock"
    LOCK_SET_USER_CODE = "lock_set_user_code"
    LOCK_ENAABLE_USER_CODE = "lock_enable_user_code"
    LOCK_DISABLE_USER_CODE = "lock_disable_user_code"
    LOCK_CLEAR_USER_CODE = "lock_clear_user_code"

    CLIMATE_SET_TEMPERATURE = "climate_set_temperature"
    CLIMATE_SET_HVAC_MODE = "climate_set_hvac_mode"
    CLIMATE_SET_FAN_MODE = "climate_set_fan_mode"
    CLIMATE_SET_PRESET_MODE = "climate_set_preset_mode"

    COVER_OPEN = "cover_open"
    COVER_CLOSE = "cover_close"
    COVER_STOP = "cover_stop"
    COVER_SET_POSITION = "cover_set_position"

    FAN_TURN_ON = "fan_turn_on"
    FAN_TURN_OFF = "fan_turn_off"
    FAN_SET_PERCENTAGE = "fan_set_percentage"
    FAN_SET_PRESET_MODE = "fan_set_preset_mode"

    BUTTON_PRESS = "button_press"

    ALARM_CONTROL_PANEL_DISARM = "alarm_control_panel_disarm"
    ALARM_CONTROL_PANEL_ARM_HOME = "alarm_control_panel_arm_home"
    ALARM_CONTROL_PANEL_ARM_AWAY = "alarm_control_panel_arm_away"
    ALARM_CONTROL_PANEL_ARM_NIGHT = "alarm_control_panel_arm_night"
    ALARM_CONTROL_PANEL_TRIGGER = "alarm_control_panel_trigger"

    SELECT_SELECT_OPTION = "select_select_option"

    NUMBER_SET_VALUE = "number_set_value"

    PLATFORM_ENTITY_REFRESH_STATE = "platform_entity_refresh_state"

    CLIENT_LISTEN = "client_listen"
    CLIENT_LISTEN_RAW_ZCL = "client_listen_raw_zcl"
    CLIENT_DISCONNECT = "client_disconnect"


class MessageTypes(StrEnum):
    """WS message types."""

    EVENT = "event"
    RESULT = "result"


class EventTypes(StrEnum):
    """WS event types."""

    CONTROLLER_EVENT = "controller_event"
    PLATFORM_ENTITY_EVENT = "platform_entity_event"
    RAW_ZCL_EVENT = "raw_zcl_event"
    DEVICE_EVENT = "device_event"


class ControllerEvents(StrEnum):
    """WS controller events."""

    DEVICE_JOINED = "device_joined"
    RAW_DEVICE_INITIALIZED = "raw_device_initialized"
    DEVICE_REMOVED = "device_removed"
    DEVICE_LEFT = "device_left"
    DEVICE_FULLY_INITIALIZED = "device_fully_initialized"
    DEVICE_CONFIGURED = "device_configured"
    GROUP_MEMBER_ADDED = "group_member_added"
    GROUP_MEMBER_REMOVED = "group_member_removed"
    GROUP_ADDED = "group_added"
    GROUP_REMOVED = "group_removed"


class PlatformEntityEvents(StrEnum):
    """WS platform entity events."""

    PLATFORM_ENTITY_STATE_CHANGED = "platform_entity_state_changed"


class RawZCLEvents(StrEnum):
    """WS raw ZCL events."""

    ATTRIBUTE_UPDATED = "attribute_updated"


ATTR_UNIQUE_ID: Final[str] = "unique_id"
COMMAND: Final[str] = "command"
CONF_BAUDRATE: Final[str] = "baudrate"
CONF_CUSTOM_QUIRKS_PATH: Final[str] = "custom_quirks_path"
CONF_DATABASE: Final[str] = "database_path"
CONF_DEFAULT_LIGHT_TRANSITION: Final[str] = "default_light_transition"
CONF_DEVICE_CONFIG: Final[str] = "device_config"
CONF_ENABLE_IDENTIFY_ON_JOIN: Final[str] = "enable_identify_on_join"
CONF_ENABLE_QUIRKS: Final[str] = "enable_quirks"
CONF_FLOWCONTROL: Final[str] = "flow_control"
CONF_RADIO_TYPE: Final[str] = "radio_type"
CONF_USB_PATH: Final[str] = "usb_path"
CONF_ZIGPY: Final[str] = "zigpy_config"

DEVICE: Final[str] = "device"

EVENT: Final[str] = "event"
EVENT_TYPE: Final[str] = "event_type"

MESSAGE_TYPE: Final[str] = "message_type"

IEEE: Final[str] = "ieee"
NWK: Final[str] = "nwk"
PAIRING_STATUS: Final[str] = "pairing_status"


DEVICES: Final[str] = "devices"
GROUPS: Final[str] = "groups"
DURATION: Final[str] = "duration"
ERROR_CODE: Final[str] = "error_code"
ERROR_MESSAGE: Final[str] = "error_message"
MESSAGE_ID: Final[str] = "message_id"
SUCCESS: Final[str] = "success"
WEBSOCKET_API: Final[str] = "websocket_api"
ZIGBEE_ERROR_CODE: Final[str] = "zigbee_error_code"


MINIMAL_MESSAGE_SCHEMA: Final = voluptuous.Schema(
    {
        voluptuous.Required(COMMAND): voluptuous.In(
            [str(item) for item in APICommands]
        ),
        voluptuous.Required(MESSAGE_ID): int,
    },
    extra=voluptuous.ALLOW_EXTRA,
)



"""WS api for the select platform entity."""
from __future__ import annotations

from typing import TYPE_CHECKING, Any

import voluptuous as vol

from zhaws.server.platforms.api import (
    execute_platform_entity_command,
    platform_entity_command_schema,
)
from zhaws.server.websocket.api import decorators, register_api_command

if TYPE_CHECKING:
    from zhaws.server.websocket.client import Client
    from zhaws.server.websocket.server import Server

ATTR_OPTION = "option"
COMMAND_SELECT_OPTION = "select_select_option"


@decorators.websocket_command(
    platform_entity_command_schema(
        COMMAND_SELECT_OPTION,
        {
            vol.Required(ATTR_OPTION): str,
        },
    )
)
@decorators.async_response
async def select_option(
    server: Server, client: Client, message: dict[str, Any]
) -> None:
    """Select an option."""
    await execute_platform_entity_command(
        server, client, message, "async_select_option"
    )


def load_api(server: Server) -> None:
    """Load the api command handlers."""
    register_api_command(server, select_option)



"""Number platform for zhawss."""
from __future__ import annotations

import functools
import logging
from typing import TYPE_CHECKING, Any

from zhaws.server.platforms import PlatformEntity
from zhaws.server.platforms.registries import PLATFORM_ENTITIES, Platform
from zhaws.server.zigbee.cluster import (
    CLUSTER_HANDLER_EVENT,
    ClusterAttributeUpdatedEvent,
)
from zhaws.server.zigbee.cluster.const import CLUSTER_HANDLER_ANALOG_OUTPUT

if TYPE_CHECKING:
    from zhaws.server.zigbee.cluster import ClusterHandler
    from zhaws.server.zigbee.device import Device
    from zhaws.server.zigbee.endpoint import Endpoint

STRICT_MATCH = functools.partial(PLATFORM_ENTITIES.strict_match, Platform.NUMBER)

_LOGGER = logging.getLogger(__name__)


@STRICT_MATCH(cluster_handler_names=CLUSTER_HANDLER_ANALOG_OUTPUT)
class Number(PlatformEntity):
    """Representation of a zhawss number."""

    PLATFORM = Platform.LOCK

    def __init__(
        self,
        unique_id: str,
        cluster_handlers: list[ClusterHandler],
        endpoint: Endpoint,
        device: Device,
    ):
        """Initialize the number."""
        super().__init__(unique_id, cluster_handlers, endpoint, device)
        self._analog_output_cluster_handler: ClusterHandler = self.cluster_handlers[
            CLUSTER_HANDLER_ANALOG_OUTPUT
        ]
        self._analog_output_cluster_handler.on_event(
            CLUSTER_HANDLER_EVENT, self._handle_event_protocol
        )

    @property
    def value(self) -> float:
        """Return the current value."""
        return self._analog_output_cluster_handler.present_value

    @property
    def min_value(self) -> float:
        """Return the minimum value."""
        min_present_value = self._analog_output_cluster_handler.min_present_value
        if min_present_value is not None:
            return min_present_value
        return 0

    @property
    def max_value(self) -> float:
        """Return the maximum value."""
        max_present_value = self._analog_output_cluster_handler.max_present_value
        if max_present_value is not None:
            return max_present_value
        return 1023

    @property
    def step(self) -> float | None:
        """Return the value step."""
        return self._analog_output_cluster_handler.resolution

    @property
    def name(self) -> str:
        """Return the name of the number entity."""
        description = self._analog_output_cluster_handler.description
        if description is not None and len(description) > 0:
            return f"{super().name} {description}"
        return super().name

    def handle_cluster_handler_attribute_updated(
        self, event: ClusterAttributeUpdatedEvent
    ) -> None:
        """Handle value update from cluster handler."""
        self.maybe_send_state_changed_event()

    async def async_set_value(self, value: Any) -> None:
        """Update the current value from service."""
        num_value = float(value)
        if await self._analog_output_cluster_handler.async_set_present_value(num_value):
            self.maybe_send_state_changed_event()

    async def async_update(self) -> None:
        """Attempt to retrieve the state of the entity."""
        await super().async_update()
        _LOGGER.debug("polling current state")
        if self._analog_output_cluster_handler:
            value = await self._analog_output_cluster_handler.get_attribute_value(
                "present_value", from_cache=False
            )
            _LOGGER.debug("read value=%s", value)

    def to_json(self) -> dict:
        json = super().to_json()
        json["engineer_units"] = self._analog_output_cluster_handler.engineering_units
        json["application_type"] = self._analog_output_cluster_handler.application_type
        json["step"] = self.step
        json["min_value"] = self.min_value
        json["max_value"] = self.max_value
        json["name"] = self.name
        return json

    def get_state(self) -> dict:
        """Return the state of the entity."""
        response = super().get_state()
        response["state"] = self.value
        return response
